/* Command Execution Module  "atom.c"           -*- C -*- */

/*
#S#
#
# This Program is part of SV (Shelx-output Viewer)
# Copyright (C) 1991-1998 Takashi Nemoto
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details. (File "COPYING")
#
#    Note!: General Public Licence is copyrighted by the Free Software 
#           Foundation,  but this code is copyrighted by me.
#
#    Send bugs and comments to tnemoto@eels.kuicr.kyoto-u.ac.jp
#
#E#
*/

#define __atom_c__
#define __cmd_c_proc__
#define __subs_c_proc__
#define __bond_c_proc__
#define __stack_c_proc__
#define __string_proc__
#define __syssub__
#include "config.h"
#ifdef Xt
#define __xtcalls_c_proc__
#define __plotxt_c_proc__
#endif
#ifdef WINDOWS
#define __plotwin_c_proc__
#endif

#include <math.h>
#include <ctype.h>
#include "def.h"

SV_Function void DecodeSfac(void)
{
  int i,j;
  for(i=0;i<n_sfac;i++) {
    if (sfac[i].name[1]==0) sfac[i].name[1]=' ';
    for(j=0;j<MAX_VDW;j++){
      if (sfac[i].name[0]==vdw[j].name[0] && 
          sfac[i].name[1]==vdw[j].name[1]) {
        sfac[i].rad=vdw[j].rad;
        sfac[i].vdw=vdw[j].vdw;
        sfac[i].col=vdw[j].col;
        break;
      }
    }
  }
}

SV_Function void del_atom(void)
{
  future();
}

SV_Function void make_atom(void)
{
  int i;
  char buf[80];
  NewAtom();
  sprintf(orgAtom[n_atom].name,"%-4s",cmd_param[0]);
  for(i=0;i<3;i++){
    orgAtom[n_atom].loc[i]=cmd_num[i+1];
  }
  sprintf(buf,"add atom %-4s %9.6f %9.6f %9.6f 11.0 0.06\n",
	  orgAtom[n_atom].name,orgAtom[n_atom].loc[0],
	  orgAtom[n_atom].loc[1],orgAtom[n_atom].loc[2]);
  n_atom++;
  updlog(buf);
  NewSat();
  setup_sym0(n_satom-1,n_atom-1);
  refresh();
}

SV_Function void rename_atom(void)
{
  ATOM_PARAM *at;
  char buf[80];
  if (cmd_atom[0]==NULL) {
    cmderr("Can not find target atom");
    return;
  } else if (cmd_atom[1]!=NULL) {
    cmderr("Dup Name");
    return;
  };
  if (cmd_c!=2) {
    cmdinfo("Usage: rename <orgname> <newname>");
    return;
  }
  at=orgAtom+cmd_atom[0]->atm;
  sprintf(at->name,"%-4s",cmd_param[1]);
  if (cmd_param[0][0]=='Q') {
    sprintf(buf,"add atom %-4s %9.6f %9.6f %9.6f 11.0 0.06\n",cmd_param[1],
           at->loc[0],at->loc[1],at->loc[2]);
  } else {
    sprintf(buf,"chg name %-4s %-4s\n",cmd_param[0],cmd_param[1]);
  }
  refresh();
}

SV_Function char *mkatmn(char *atm)
{
  static char atmn[5];
  int max=1000;
  int i;
  if (atm[1]!=0) max=100;
  for(i=1;i<max;i++) {
    sprintf(atmn,"%s%d",atm,i);
    if (getatm(atmn)==NULL) return atmn;
  }
  sprintf(atmn,"Err");
  return atmn;
}

SV_Function void name(void)
{
  int i,j;
  ATOM_PARAM *at;
  char atm[3];
  char *atmn;
  char buf[80];
  if (cmd_c<2) {cmdinfo("Usage: name <org name> <new name>");return;}
  atm[2]=0;
  atm[0]=cmd_param[0][0];
  atm[1]=cmd_param[0][1];
  if (atm[1]!=0 && !isalpha(atm[1])) {cmderr("## name: name err");return;}
  if (cmd_c==2 && chkinst(cmd_param[1],"SL","SELECT")) {
    cmd_param[1][0]=0;
    global_flag|=GLOBALFLAG_NOTREFRESH;
    while(NULL!=(cmd_atom[1]=popstack())) {
      name();
    }
    global_flag&=~GLOBALFLAG_NOTREFRESH;
    refresh();
    return;
  }
  for(i=1;i<cmd_c;i++) {
    at=orgAtom+cmd_atom[i]->atm;
    atmn=mkatmn(atm);
    if (cmd_param[i][0]=='Q') {
      sprintf(buf,"add atom %-4s %9.6f %9.6f %9.6f 11.0 0.06\n",atmn,
              at->loc[0],at->loc[1],at->loc[2]);
    } else {
      sprintf(buf,"chg name %-4s %-4s\n",cmd_param[i],mkatmn(atm));
    } 
    sprintf(at->name,"%s",atmn);
    if (atm[1]==0) atm[1]=32;
    for(j=0;j<n_sfac;j++){
      if (strcmp(sfac[j].name,atm)==0) {
	break;
      }
    }
    if (j==n_sfac) {
      strcpy(sfac[n_sfac].name,atm);
      n_sfac++;
      DecodeSfac();
    }
    at->atom_type=j;
    updlog(buf);
  }
  for(i=0;i<n_satom;i++){
    SAT[i].color=sfac[OrgAtom(i)->atom_type].col;
  }
  refresh();
}

SV_Function void build_on(void)
{
  int i,j;
  int f;
  int nFrag=1;
  int *BondList;
  int *nBondList;
  int **BondListTag;
  BondList=(int *)Calloc(n_satom*MAX_BOND_PAR_ATOM,sizeof(int));
  nBondList=(int *)Calloc(n_satom,sizeof(int));
  BondListTag=(int **)Calloc(n_satom,sizeof(int *));
  for(i=0;i<n_satom;i++) {
    BondListTag[i]=BondList+i*MAX_BOND_PAR_ATOM;
    nBondList[i]=0;
    SAT[i].temp_num=i+1;
  }
  if (cmd_c==1 && cmd_atom[0]!=NULL) {
    cmd_atom[0]->temp_num=0;
  } else {
    SAT[0].temp_num=0;
  }
  for(i=0;i<n_bond;i++){
    int at1,at2;
    at1=BOND[i].atm[0];
    at2=BOND[i].atm[1];
    if ((SAT[at1].view_flag & FLAG_OMIT) ==0) {
      BondListTag[at1][nBondList[at1]]=at2;
      BondListTag[at2][nBondList[at2]]=at1;
      nBondList[at1]++;
      nBondList[at2]++;
    }
  }
  do {
    f=0;
    for(i=0;i<n_satom;i++) {
      for(j=nBondList[i]-1;j>=0;j--) {
	if (SAT[BondListTag[i][j]].temp_num > SAT[i].temp_num) {
	  SAT[BondListTag[i][j]].temp_num=SAT[i].temp_num;
	  f=1;
	} else if (SAT[BondListTag[i][j]].temp_num < SAT[i].temp_num) {
	  SAT[i].temp_num=SAT[BondListTag[i][j]].temp_num;
	  f=1;
	}
      }
    }
  } while (f!=0);
  for(i=0;i<n_satom;i++){
    nBondList[i]=-1;
  }
  nBondList[0]=0;
  nFrag=1;
  for(i=0;i<n_satom;i++){
    nBondList[i]=-1;
  }
  for(i=0;i<n_satom;i++){
    if (nBondList[SAT[i].temp_num]<0) {
      SAT[i].frag=nFrag;
      nBondList[SAT[i].temp_num]=nFrag++;
    } else {
      SAT[i].frag=nBondList[SAT[i].temp_num];
    }
  }
  
  Free(BondListTag);
  Free(nBondList);
  Free(BondList);
}
    
SV_Function Double get_ueq(ATOM_PARAM *at)
{
  Double ueq;
  int i;
  ueq=0;
  for(i=0;i<3;i++)
    ueq+=((at->uij[i])*cell[i]*cell[i]*rcell[i]*rcell[i]
      +(at->uij[i+3])*2*cell[i+1<3?i+1:i-2]*cell[i+2<3?i+2:i-1]
          *rcell[i+1<3?i+1:i-2]*rcell[i+2<3?i+2:i-1]*cell_cos[i]);
  return ueq/3;
}

SV_Function void dispxyz(void)
{
  int i;
  S_AT *at1;
  Double *nloc,len;
  static S_AT *pa=NULL;
  ATOM_PARAM *oat;
  sprintf(disp_str[0],"XYZ ");
  for(i=0;i<cmd_c;i++){
    if (i>0) pa=cmd_atom[1];
    at1=cmd_atom[i];
    if (at1==NULL) break;
    nloc=sym_exp(at1,orgAtom);
    sprintf(disp_str[i+1],"%-4s %8.5f %8.5f %8.5f",orgAtom[at1->atm].name,
            nloc[0],nloc[1],nloc[2]);
    sprintf(disp_str[i+1],"SYMM%2d %3d%3d%3d",at1->sym,
            at1->sft[0],at1->sft[1],at1->sft[2]);
    sprintf(disp_str[i+1],"Cart loc %8.5f %8.5f %8.5f",
            at1->vloc[0],at1->vloc[1],at1->vloc[2]);
  }
  at1=cmd_atom[0];
  if (at1==NULL) {
    disp_str[0][0]=0;
    refresh();
    return;
  }
  oat=orgAtom+at1->atm;
  if (at1!=NULL){
    i++;
    sprintf(disp_str[i++]," ");
    sprintf(disp_str[i++],"Information of %s ",oat->name);
    sprintf(disp_str[i++],"OrgLoc. %9.5f%9.5f%9.5f",
	    oat->loc[0],oat->loc[1],oat->loc[2]);
    sprintf(disp_str[i++],"Occ. %9.5f ",oat->occ);
    if (oat->peak_h!=0) {
      sprintf(disp_str[i++],"Peak Height %9.5f ",oat->peak_h);
    } else {
      if (oat->uij[1]==0) {
        sprintf(disp_str[i++],"Uiso=%9.5f    Biso=%9.5f  ",
		oat->uij[0],oat->uij[0]*8*PI*PI);
      } else {
        sprintf(disp_str[i++],"Ueq =%9.5f    Beq =%9.5f  ",
                get_ueq(oat),get_ueq(oat)*8*PI*PI);
        sprintf(disp_str[i++],"U11       U22       U33");
        sprintf(disp_str[i++],"%9.5f %9.5f %9.5f",
                oat->uij[0],oat->uij[1],oat->uij[2]);
        sprintf(disp_str[i++],"U23       U13       U12");
        sprintf(disp_str[i++],"%9.5f %9.5f %9.5f",
                oat->uij[3],oat->uij[4],oat->uij[5]);
      }
    }
    sprintf(disp_str[i++],"AFIX   PART   FRAG");
    sprintf(disp_str[i++],"%4d   %4d   %4d",
                oat->afix,oat->part,at1->frag);
    if (at1->sym!=0 || at1->sft[0]!=0 || at1->sft[1]!=0 || at1->sft[2]!=0) {
      sprintf(disp_str[i++]," ");
      sprintf(disp_str[i++],"%-4s with Symmetry Op",oat->name);
      nloc=sym_exp(at1,orgAtom);
      sprintf(disp_str[i++],"%9.5f %9.5f %9.5f",nloc[0],nloc[1],nloc[2]);
    }
  }
  sprintf(disp_str[i++]," ");
  if (IsNonPosi(oat->uij)){
    sprintf(disp_str[i++],"Warning!! Non Positive Temp Factor");
    sprintf(disp_str[i++]," ");
  }  
  if (pa!=NULL) {
    int j;
    Double x1[3],x2[3],x3[3],*x4;
    len=get_length(pa,at1);
    x4=sym_exp(at1,orgAtom);
    x3[0]=x4[0];x3[1]=x4[1];x3[2]=x4[2];
    x4=sym_exp(pa,orgAtom);
    if (len>2 && len<4) {
      for(j=0;j<3;j++) { 
        x1[j]=orgAtom[at1->atm].loc[j];
        x2[j]=x3[j]+(x4[j]-x3[j])/len*1.0;
        x1[j]=x4[j]+(x3[j]-x4[j])/len*1.0;
      }
      sprintf(disp_str[i++],"Hydrogen bond");
      sprintf(disp_str[i++],"%-4s  H1   H2  %-4s",
              orgAtom[at1->atm].name,orgAtom[pa->atm].name);
      sprintf(disp_str[i++]," -1.00-     -1.00-");
      sprintf(disp_str[i++],"H1 %8.5f %8.5f %8.5f ",x2[0],x2[1],x2[2]);
      sprintf(disp_str[i++],"H2 %8.5f %8.5f %8.5f ",x1[0],x1[1],x1[2]);
    }
  }
  pa=at1;
  disp_str[i][0]=0;
  refresh();
  return;
}

SV_Function void unpack(void)
{
  unpack_sub(0);
  n_bond=0;
  InsertCommandQueue("makebond auto");
}

SV_Function void unpack_sub(int omit_q)
{
  int i;
  for(i=0;i<n_atom;i++) {
    orgAtom[i].temp_num=(n_satom==0?
                      (orgAtom[i].name[0]=='Q'?FLAG_OMIT*omit_q:0)
                      :(FLAG_OMIT | FLAG_NOLABEL | FLAG_TEMP));
  }
  for(i=0;i<n_satom;i++) {
    OrgAtom(i)->temp_num&=SAT[i].view_flag;
  };
  n_satom=0;
  for(i=0;i<n_atom;i++) {
    SAT[i].num=i;
    SAT[i].atm=i;
  }
  for(i=0;i<n_atom;i++) {
    if ((orgAtom[i].temp_num & FLAG_TEMP)!=0){
      orgAtom[i].temp_num=FLAG_OMIT;
    };
  }
  for (n_satom=0;n_satom<n_atom;n_satom++){
    SAT[n_satom].view_flag=orgAtom[n_satom].temp_num;
    setup_sym0(n_satom,n_satom);
    SAT[n_satom].frag=orgAtom[n_satom].frag;
  };
}

SV_Function void pack(void)
{
  Double xmin,xmax,ymin,ymax,zmin,zmax;
  Double loc[3],shift[3],*lx,x,y,z,x0,y0,z0;
  S_AT sat,*satp;
  int i,j,k,l,flag;
  int n_satom0;
  if (cmd_c!=0 && cmd_c!=6 && cmd_c!=1 && cmd_c!=2) {
    cmdinfo("usage: pack [x0 x1 y0 y1 z0 z1]");
    return;
  };
  if (cmd_c==6) {
    xmin=cmd_num[0];  xmax=cmd_num[1];
    ymin=cmd_num[2];  ymax=cmd_num[3];
    zmin=cmd_num[4];  zmax=cmd_num[5];
  } else if (cmd_c==1) {
    xmin=ymin=zmin=0-cmd_num[0];
    xmax=ymax=zmax=1+cmd_num[0];
  } else if (cmd_c==2) {
    xmin=ymin=zmin=cmd_num[0];
    xmax=ymax=zmax=cmd_num[1];
  } else {
    xmin=-0.1;  xmax=1.1;
    ymin=-0.1;  ymax=1.1;
    zmin=-0.1;  zmax=1.1;
  }
#ifdef DEBUG
  fprintf(stderr,"%f %f %f %f %f %f\n",xmin,xmax,ymin,ymax,zmin,zmax);
#endif
  if (xmin>xmax) {
      cmderr("pack: invalid range");
  }
  atom_stack_top=atom_stack_top_bak=-1;
  for(i=0;i<n_atom;i++) {
    orgAtom[i].temp_num=FLAG_OMIT | FLAG_NOLABEL;
  }
  for(i=0;i<n_satom;i++) {
    OrgAtom(i)->temp_num&=SAT[i].view_flag;
  };
  n_satom=0;
  sat.sft[0]=sat.sft[1]=sat.sft[2]=0;
  dispgraph(NULL,-1);
  for(i=0;i<n_atom;i++){
    dispgraph("Make Atoms",i*100/n_atom);
    n_satom0=n_satom;
    sat.atm=i;
    if ((orgAtom[i].temp_num & FLAG_OMIT)!=0) continue;
    for(j=0;j<symnum;j++) {
      sat.sym=j;
      lx=sym_exp(&sat,orgAtom);
      for(k=0;k<3;k++) {
        shift[k]=floor(lx[k]);
        loc[k]=lx[k]-shift[k]; /*  0<=loc[k]<1  */
      };
      for(x=floor(xmin);x<=xmax;x++) {
        if (x+loc[0]>=xmin && x+loc[0]<=xmax) {
          for(y=floor(ymin);y<=ymax;y++) {
            if (y+loc[1]>=ymin && y+loc[1]<=ymax) {
              for(z=floor(zmin);z<=zmax;z++) {
                if (z+loc[2]>=zmin && z+loc[2]<=zmax) {
/*
                  if (n_satom>MAX_SYMATOM*0.9) {
                      SAT=Realloc(SAT,&MAX_SYMATOM,ATOM_BLOCK_SIZE,
				  sizeof(S_AT));
                      if (SAT==NULL) {
                          PutMessage1("Out of Memory!");
                          Exit(1);
                      }
                  }
*/
                  satp=NewSat();
                  satp->atm=i;
                  satp->sym=j;
                  satp->sft[0]=x-shift[0];
                  satp->sft[1]=y-shift[1];
                  satp->sft[2]=z-shift[2];
                  satp->view_flag=orgAtom[i].temp_num;
                  satp->frag=orgAtom[i].frag;
                  satp->part=orgAtom[i].part;
                  setup_symloc(satp,orgAtom);
                  flag=0;
                  for(l=n_satom0;l<n_satom-1;l++){
                    x0=satp->vloc[0]-SAT[l].vloc[0];
                    y0=satp->vloc[1]-SAT[l].vloc[1];
                    z0=satp->vloc[2]-SAT[l].vloc[2];
                    if (x0*x0+y0*y0+z0*z0<0.01) {
                      flag=1;break;
                    }
                  }
                  if (flag==0){
                    setup_atom(satp);
                  } else {
                    n_satom--;
                  }
                }
              }
            }
          }
        }
      }
    }    
  }
  dispgraph(NULL,-2);
  InsertCommandQueue("makebond auto");
}

SV_Function void pack2(void)
{
  Double *loc;
  int i;
  int *inCell;
  Float xxx;
  if (cmd_c>1) {
    cmdinfo("usage: pack2 [num]");
    return;
  }
  inCell=(int *)Calloc(n_satom,sizeof(int));
  if (cmd_c==1) xxx=cmd_num[0]; else xxx=0;
  cmd_c=1;
  cmd_atom[0]=NULL;
  global_flag|=GLOBALFLAG_NOTREFRESH;
  build_on();
  global_flag&=~GLOBALFLAG_NOTREFRESH;
/*  for(i=0;i<500;i++)inCell[i]=0;*/
  for(i=0;i<n_satom;i++){
    loc=sym_exp(SAT+i,orgAtom);
    if (-xxx<=loc[0] && loc[0]<=1+xxx &&
        -xxx<=loc[1] && loc[1]<=1+xxx &&
        -xxx<=loc[2] && loc[2]<=1+xxx) {
      inCell[SAT[i].frag]=1;
    }
  }
  for(i=0;i<n_satom;i++){
    if (inCell[SAT[i].frag]==0) SAT[i].view_flag |= FLAG_OMIT;
    if (SAT[i].frag==0) {
      loc=sym_exp(SAT+i,orgAtom);
      if ( ! (-xxx<=loc[0] && loc[0]<=1+xxx &&
              -xxx<=loc[1] && loc[1]<=1+xxx &&
              -xxx<=loc[2] && loc[2]<=1+xxx)) {
        SAT[i].view_flag |= FLAG_OMIT;
      }
    }
  }
  Free(inCell);
  refresh();
}


SV_Function void setatomcolor(void)
{
  int i;
  if (cmd_c<2) {cmderr("Usage: color <atom name> <color>");return;}
  if (cmd_c==2 && chkinst(cmd_param[1],"SL","SELECT")) {
    cmd_param[1][0]=0;
    global_flag|=GLOBALFLAG_NOTREFRESH;
    while(NULL!=(cmd_atom[1]=popstack())) {
      setatomcolor();
    }
    global_flag&=~GLOBALFLAG_NOTREFRESH;
    refresh();
    return;
  }
  for(i=1;i<cmd_c;i++) {
    cmd_atom[i]->color=cmd_int[0];
  }
  refresh();
}

