/* Cavity Interface "cavity.c"   -*- C -*- */

/*
#S#
#
# This Program is part of SV (Shelx-output Viewer)
# Copyright (C) 1991-1998 Takashi Nemoto
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details. (File "COPYING")
#
#    Note!: General Public Licence is copyrighted by the Free Software 
#           Foundation,  but this code is copyrighted by me.
#
#    Send bugs and comments to tnemoto@eels.kuicr.kyoto-u.ac.jp
#
#E#
*/

#define __cavity_c__
#define __cmd_c_proc__
#define __subs_c_proc__
#define __stack_c_proc__
#define __cavmenu_c_proc__
#define __readf_c_proc__
#define __readmeta_c_proc__
#define __string_proc__

#ifdef Xt
#define __xtcalls_c_proc__
#define __plotxt_c_proc__
#endif

#include "def.h"
#include <ctype.h>
#include <math.h>
#ifdef CAVITY_EXEC
#ifdef __UNIX__
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <signal.h>
#endif
#endif
#define MAX_CAV 50000
#define MAX_CAVATOM 3000
#define MAX_CAVBOND 6000
#define MAX_CAVEXCEPT 3000
#define MAX_CAVRADII 20

SV_Variable Double min[3],max[3],vmin[3],vmax[3],vorg[3],*loc,RVMT[3][3];
SV_Variable Double cav_radii[MAX_CAVRADII];
SV_Variable Float cavminmax[6]
                           ={0,0,0,0,0,0};
SV_Variable int cavitymode[10];

SV_Function void cavdraw(void)
{
  int imin[3],imax[3],cav_except[MAX_CAVEXCEPT];
  int i,j,n,n_cav,n_cav2,cav_bond[MAX_CAVBOND],cav_atom[MAX_CAVATOM];
  int n_cavatom,n_cavexcept,n_cavbond,tmp1,tmp2;
  int auto_flag=0,vol_flag=1,bond_flag=1,exec_flag=1;
  Float atom_radi=0,grid=0.2;
  Mini loc0[3],k,l,m,vw_flag;
  S_AT sat,*satp,*satp2,*satp3;
  S_AT *CAT;  /* atom for cavity */
  FILE *cav_fp,*cavcnt_fp;
/* initialize */
  bond_flag=0;
  vw_flag=0;
  if (cmd_c==0 || cmd_int[0]==0) vw_flag=3;
  if (cmd_c>1) auto_flag=cmd_int[1];
  if (cmd_c>2) vol_flag=cmd_int[2];
  if (cmd_c>3) bond_flag=cmd_int[3];
  if (cmd_c>4) atom_radi=cmd_num[4];
  if (cmd_c>5) grid=cmd_num[5];
  if (cmd_c>6) exec_flag=cmd_int[6];
  if (atom_stack_top<0) {
    cmderr("Cavity Atom Empty!!");
    return;
  }
#ifdef Xt
  if (cmd_c==0 && X11MODE) {
    cavmenu_popup();
    return;
  };
#endif
  CAT=(S_AT *)Malloc(sizeof(S_AT)*MAX_CAV);
  if (CAT==NULL) {
    cmderr("Out of Memory");
    return;
  }
  for(i=0;i<3;i++) {
    min[i]=10000; max[i]=-10000;
    vmin[i]=10000; vmax[i]=-10000;
  }
  for(i=0;i<=atom_stack_top;i++) {
    satp=SAT+atom_stack[i];
    loc=sym_exp(satp,orgAtom);
    for(j=0;j<3;j++) {
      if (loc[j]<min[j]) min[j]=loc[j];
      if (loc[j]>max[j]) max[j]=loc[j];
      if (satp->vloc[j]<vmin[j]) vmin[j]=satp->vloc[j];
      if (satp->vloc[j]>vmax[j]) vmax[j]=satp->vloc[j];
    }
  }
  for(i=0;i<3;i++){
    imin[i]=floor(min[i])-1;
    imax[i]=floor(max[i])+2;
    vmin[i]-=6;
    vmax[i]+=6;
  }
  n_cav=n_cav2=0;
  for(i=0;i<n_atom;i++){
    sat.atm=i;
    if (orgAtom[i].name[0]=='Q') continue;
    for(j=0;j<symnum;j++){
      sat.sym=j;
      sat.sft[0]=0;
      sat.sft[1]=0;
      sat.sft[2]=0;
      loc=sym_exp(& sat,orgAtom);
      for(k=0;k<3;k++) loc0[k]=-floor(loc[k]);
      for(k=imin[0];k<=imax[0];k++){
	sat.sft[0]=loc0[0]+k;
	for(l=imin[1];l<=imax[1];l++){
	  sat.sft[1]=loc0[1]+l;
	  for(m=imin[2];m<=imax[2];m++){
	    sat.sft[2]=loc0[2]+m;
	    setup_symloc(& sat,orgAtom);
	    if (sat.vloc[0]<vmin[0] || sat.vloc[0]>vmax[0]) continue;
	    if (sat.vloc[1]<vmin[1] || sat.vloc[1]>vmax[1]) continue;
	    if (sat.vloc[2]<vmin[2] || sat.vloc[2]>vmax[2]) continue;
	    satp=CAT+n_cav;
	    satp->atm=i;
	    satp->sym=j;
	    satp->temp_num=n_cav;
	    satp->num=n_cav2;
	    satp->sft[0]=k+loc0[0];
	    satp->sft[1]=l+loc0[1];
	    satp->sft[2]=m+loc0[2];
            setup_symloc(satp,orgAtom);
	    n_cav++;	
	    n_cav2++;
	    for(n=0;n<n_cav-1;n++){
	      satp2=CAT+n;
	      if (((satp->vloc[0]-satp2->vloc[0])*
		   (satp->vloc[0]-satp2->vloc[0])+
		   (satp->vloc[1]-satp2->vloc[1])*
		   (satp->vloc[1]-satp2->vloc[1])+
		   (satp->vloc[2]-satp2->vloc[2])*
		   (satp->vloc[2]-satp2->vloc[2])) < 0.0001) {
		n_cav2--;
		satp->temp_num=n;
		break;
	      }
	    }
	  }
	}
      }
    }
  }
/* File Open */
  cav_fp=ffopen("cav.dat","w");
  if (cav_fp==NULL) {
    cmderr("Can not open cav.dat.");
    Free(CAT);
    return;
  }
  cavcnt_fp=ffopen("cav.cnt","w");
  if (cavcnt_fp==NULL) {
    cmderr("Can not open cav.cnt.");
    fclose(cav_fp);
    Free(CAT);
    return;
  }
  n_cavatom=n_cavbond=n_cavexcept=0;
  for(i=0;i<n_bond;i++){
    satp=SAT+BOND[i].atm[0];
    satp2=SAT+BOND[i].atm[1];
    tmp1=tmp2=-1;
    if (((satp->view_flag & FLAG_OMIT)==0 &&
	(satp2->view_flag & FLAG_OMIT)==0)) {
      for(j=0;j<n_cav;j++){
	satp3=CAT+j;
	if (satp->atm==satp3->atm && satp->sym==satp3->sym &&
	    satp->sft[0]==satp3->sft[0] && satp->sft[1]==satp3->sft[1] &&
	    satp->sft[2]==satp3->sft[2]) {
	  tmp1=satp3->num;
	}
	if (satp2->atm==satp3->atm && satp2->sym==satp3->sym &&
	    satp2->sft[0]==satp3->sft[0] && satp2->sft[1]==satp3->sft[1] &&
	    satp2->sft[2]==satp3->sft[2]) {
	  tmp2=satp3->num;
	}
      }
      if (tmp1>=0 && tmp2>=0) {
        cav_bond[n_cavbond++]=tmp1;
        cav_bond[n_cavbond++]=tmp2;
      }
    }
  }
  for(i=0;i<n_cav;i++){
    satp=CAT+i;
    for(j=0;j<=atom_stack_top;j++){
      satp3=SAT+atom_stack[j];
	if (satp->atm==satp3->atm && satp->sym==satp3->sym &&
	    satp->sft[0]==satp3->sft[0] && satp->sft[1]==satp3->sft[1] &&
	    satp->sft[2]==satp3->sft[2]) {
	  cav_except[n_cavexcept++]=satp->num;
	  break;
	}
    }
    for(j=0;j<n_satom;j++){
      satp3=SAT+j;
      if ((satp3->view_flag & FLAG_OMIT)!=0) continue;
      if (satp->atm==satp3->atm && satp->sym==satp3->sym &&
          satp->sft[0]==satp3->sft[0] && satp->sft[1]==satp3->sft[1] &&
          satp->sft[2]==satp3->sft[2]) {
	cav_atom[n_cavatom++]=satp->num;
	break;
      }
    }
  }
  if (auto_flag!=0) {
    for(i=0;i<3;i++) {
      vmin[i]=cavminmax[i*2];
      vmax[i]=cavminmax[i*2+1];
    };
  }
  fprintf(cavcnt_fp,"# Title\n");
  fprintf(cavcnt_fp,"%s\n",Title);
  fprintf(cavcnt_fp,"# iview,  Manu,Volume,  Bond,  Test, ICAVE,N_Atom,"
	  "N_Bond, N_Cav\n");
  fprintf(cavcnt_fp,"%7d%7d%7d%7d      0      0%7d%7d%7d\n",vw_flag,
	  auto_flag,vol_flag,bond_flag,n_cavatom,n_cavbond/2,n_cavexcept);
  fprintf(cavcnt_fp,"# Scale,Radius,CavPen,BondPn,CavBnP,CavAtP\n");
  fprintf(cavcnt_fp,"    1.0%7.3f      1      2      0      0\n",atom_radi);
  fprintf(cavcnt_fp,"# Atom Number\n");
  j=0;
  for(i=0;i<n_cavatom;i++){
    fprintf(cavcnt_fp,"%7d",cav_atom[i]+1);
    if (++j==10) {
      fprintf(cavcnt_fp,"\n");
      j=0;
    }
  }
  if (j!=0) { 
    fprintf(cavcnt_fp,"\n");
    j=0;
  }
  fprintf(cavcnt_fp,"# Bond Atom Number (Pair)\n");

  for(i=0;i<n_cavbond;i++){
    fprintf(cavcnt_fp,"%7d",cav_bond[i]+1);
    if (++j==10) {
      fprintf(cavcnt_fp,"\n");
      j=0;
    }
  }
  if (j!=0) { 
    fprintf(cavcnt_fp,"\n");
    j=0;
  }

  fprintf(cavcnt_fp,"# Cavity Atom Number\n");
  for(i=0;i<n_cavexcept;i++){
    fprintf(cavcnt_fp,"%7d",cav_except[i]+1);
    if (++j==10) {
      fprintf(cavcnt_fp,"\n");
      j=0;
    }
  }
  if (j!=0) { 
    fprintf(cavcnt_fp,"\n");
    j=0;
  }
  if (auto_flag==0){
    for(i=0;i<3;i++) {
      vmin[i]=10000; vmax[i]=-10000;
    }
    for(i=0;i<n_cavexcept;i++) {
      loc=CAT[cav_except[i]].vloc;
      for(j=0;j<3;j++) {
        if (loc[j]<vmin[j]) vmin[j]=loc[j];
        if (loc[j]>vmax[j]) vmax[j]=loc[j];
      }
    }
    for(i=0;i<3;i++) {
      vorg[i]=avg[i];
      vmin[i]+=-vorg[i]-3.0;
      vmax[i]+=-vorg[i]+3.0;
    }
  } else {
    vorg[0]=vorg[1]=vorg[2]=0.0;
    if (n_cavexcept>0) {
      for(i=0;i<3;i++){
	vorg[i]=CAT[cav_except[0]].vloc[i];
	vmin[i]+=-vorg[i]-3.0;
	vmax[i]+=-vorg[i]+3.0;
      }
    }
  }

  for(i=0;i<MAX_CAVRADII;i++){cav_radii[i]=-1;}
  for(i=0;i<n_sfac;i++){ cav_radii[i+1]=sfac[i].vdw; }
  l=n_sfac;
/*
  l=0;
  for(i=0;i<n_satom;i++){
    k=SAT[i].atm->atom_type;
    if (k>l) l=k;
    if (cav_radii[k]<0) {
      for(j=0;j<MAX_VDW;j++){
        if (vdw[j].name[0]==SAT[i].atm->name[0] && (
	    (!isalpha(vdw[j].name[1]) && !isalpha(SAT[i].atm->name[1])) 
	    || vdw[j].name[1]==SAT[i].atm->name[1])) {
          cav_radii[k]=vdw[j].vdw;
	}
      }
    }
  }
*/
  fprintf(cavcnt_fp,"# Cavity Radi\n");
  for(i=1;i<=l;i++){
    fprintf(cavcnt_fp,"%7.3f",cav_radii[i]+1.2);
  };
  fprintf(cavcnt_fp,"\n");
  fprintf(cavcnt_fp,"# Cavity Range\n");
  fprintf(cavcnt_fp,"# X Min, X Max, Y Min, Y Max, Z Min, Z Max, Grid\n");
  fprintf(cavcnt_fp,"%7.3f%7.3f%7.3f%7.3f%7.3f%7.3f%7.3f\n",
	  vmin[0],vmax[0],vmin[1],vmax[1],vmin[2],vmax[2],grid);
  fclose(cavcnt_fp);
  inv_matrix(VMT,RVMT);
  fprintf(cav_fp,"ORIGIN    %10.6f%10.6f%10.6f\n",vorg[0],vorg[1],vorg[2]);
  for(i=0;i<3;i++){
    fprintf(cav_fp,"VMT%1d      %10.6f%10.6f%10.6f\n",
	    i,VMT[i][0],VMT[i][1],VMT[i][2]);
  };
  for(i=0;i<3;i++){
    fprintf(cav_fp,"RVMT%1d     %10.6f%10.6f%10.6f\n",
	    i,RVMT[i][0],RVMT[i][1],RVMT[i][2]);
  };
  for(i=0;i<n_cav;i++){
    if (CAT[i].temp_num==i) {
      fprintf(cav_fp,"%-6s%4d%10.6f%10.6f%10.6f\n",
	      orgAtom[CAT[i].atm].name,orgAtom[CAT[i].atm].atom_type,
	      CAT[i].vloc[0],CAT[i].vloc[1],CAT[i].vloc[2]);
    }
  }
  fclose(cav_fp);
  Free(CAT);
#ifdef CAVITY_EXEC
  if (exec_flag!=0) exec_cav();
#endif
  return;
}     

#ifdef CAVITY_EXEC

#ifdef CAVITY_DISP_PROGRESS

static int checkpid(int pid)
{
#ifdef HAVE_WAITPID
  int waitstatus;
  if (waitpid(pid,&waitstatus,WNOHANG)==pid) return 1;
  return 0;
#else
  /*  union wait waitstatus;*/
  int waitstatus;
  if (wait3(&waitstatus,WNOHANG,0)==pid) return 1;
  return 0;
#endif
}

static void xtflush(void)
{
#ifdef Xt
  if (X11MODE)xtFlush();
#endif
}

SV_Function void waitstart(char *mess1,char *mess2,int fl,int *w_flag)
{
  if (X11MODE) {
#ifdef Xt
    WaitWin(mess1,mess2,fl,w_flag);
#endif
  } else {
    cmderr(mess1);
  }
}

SV_Function void waitend(void)
{
  if (X11MODE) {
#ifdef Xt
    WaitEnd();
#endif
  }
}

SV_Function void waitmess(char *mess)
{
  if (X11MODE) {
#ifdef Xt
    WaitMess(mess);
#endif
  } else {
    cmderr(mess);
  }
}
#endif

SV_Function void exec_cav(void)
{ 
  FILE *cav_fp;
#ifdef CAVITY_DISP_PROGRESS
  int pid,w_flag,endflag=0;
  time_t tm;
  time(&tm);
  unlink(filename(curwd,"cav.ps"));
  if  ((pid=fork())<0) {
    cmderr("Can not exec CAVITY.");
    return;
  }
  if (pid==0) {
    sprintf(fbuf,"cd %s;cav5",curwd);
    system(fbuf);
    exit(1);
  }
  waitstart("Cavity Execute.  Wait!!","Waiting Cavity Startup",1,&w_flag);
  for(;;){
    xtflush();
    if (w_flag!=0){
      cmderr("User Interrupt for CAVITY");
      xtflush();
      kill(pid,SIGKILL);
    };
    if(checkpid(pid)!=0) {
      waitend();
      return;
    }
    if ((cav_fp=ffopen("cav.ps","r"))!=0) break;
  }
  for(;;){
    if (w_flag!=0){
      cmderr("User Interrupt for CAVITY");
      xtflush();
      kill(pid,SIGKILL);
    };
#else
  {
    char tmpcurwd[1024];
    getcwd(tmpcurwd,1000);
    chdir(curwd);
#ifdef WINDOWS
    system("cavmenu"); 
#else
    system("cav5");
#endif
    chdir(tmpcurwd);
  }  
  if ((cav_fp=ffopen("cav.ps","r"))==0) {
    cmderr("Cavity: Can't open result file!");
  } else {
#endif
    while (ffgets(cav_fp)!=0) {
      if (strncmp("%# XRANGE =",fbuf,11)==0) {
	sscanf(fbuf+17,"%10lf%10lf",cavminmax,cavminmax+1);
      } 
      if (strncmp("%# YRANGE =",fbuf,11)==0) {
	sscanf(fbuf+17,"%10lf%10lf",cavminmax+2,cavminmax+3);
      } 
      if (strncmp("%# ZRANGE =",fbuf,11)==0) {
	sscanf(fbuf+17,"%10lf%10lf",cavminmax+4,cavminmax+5);
      } 
#ifdef CAVITY_DISP_PROGRESS
      if (strncmp("%# mess Z ",fbuf,10)==0) {
	waitmess(fbuf+8);
      }
#endif
      if (strncmp("%# mess Volume ",fbuf,15)==0) {
	cmdinfo(fbuf+8);
      }
    } 
#ifdef CAVITY_DISP_PROGRESS
    if (endflag!=0) {
      waitend();
      fclose(cav_fp);
      cmd_c=1;
      sprintf(cmd_param2[0],"%s","cav.ps");
      readmeta();
      return;
    }
    if(checkpid(pid)!=0) {
      endflag=1;
    }
    xtflush();
    sleep(1);
    fseek(cav_fp,0,1);
#else 
    fclose(cav_fp);
    cmd_c=1;
    sprintf(cmd_param2[0],"%s","cav.ps");
    readmeta();
#endif
  }
}

#endif /* CAVITY_EXEC */

