/* Command Execution Module  "cmd.c"           -*- C -*- */

/*
#S#
#
# This Program is part of SV (Shelx-output Viewer)
# Copyright (C) 1991-1998 Takashi Nemoto
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details. (File "COPYING")
#
#    Note!: General Public Licence is copyrighted by the Free Software 
#           Foundation,  but this code is copyrighted by me.
#
#    Send bugs and comments to tnemoto@eels.kuicr.kyoto-u.ac.jp
#
#E#
*/

#define __cmd_c__
#define __file_c_proc__
#define __svx_c_proc__
#define __subs_c_proc__
#define __view_c_proc__
#define __atom_c_proc__
#define __bond_c_proc__
#define __readmeta_c_proc__
#define __stack_c_proc__
#define __panic_c_proc__
#define __ortep_c_proc__
#define __charon_c_proc__
#define __povrayc_proc__
#define __cavity_c_proc__
#define __shelplot_c_proc__
#define __shelview_c_proc__
#ifdef Xt
#define __xtcalls_c_proc__
#endif

#include <math.h>
#include <ctype.h>
#define __syssub__
#define __string_proc__
#include "def.h"

SV_Variable char cmd_param[MAX_CMD_PARAM+1][81];
SV_Variable char cmd_param2[MAX_CMD_PARAM+1][81];

SV_Variable S_AT *atom_stack_s[10];
SV_Variable Double atom_st_bnd[9];
SV_Variable Double atom_st_ang[8];
SV_Variable Double atom_st_tor[7];

/* SV_Variable int cmd_flag; */
SV_Variable Double rot_step;
SV_Variable char BASENAME[80];
SV_Variable S_AT *cmd_atom[MAX_CMD_PARAM+1];
SV_Variable Double cmd_num[MAX_CMD_PARAM+1];
SV_Variable int cmd_int[MAX_CMD_PARAM+1];
SV_Variable int cmd_c;

char commandQueue[cBufSize][200];
int nCommandQueue=0;

struct {
  char cmd1[20];
  char cmd2[20];
  char shortHelp[80];
  void (*func)(void);
} command[]
={
  {"open","openfile","open <filename>",open_file},
  {"pop","popstack","pop (no arg)",pop_stack},
  {"cm","macro","macro <filename>",macro},
  {"mt","readmeta","readmeta <filename>",readmeta},
  {"rt","rotate","rt <rotx> <roty> <rotz> (radian)",rot},
  {"x","rotatex","x <rot> (degree)",rot_x},
  {"y","rotatey","y <rot> (degree)",rot_y},
  {"z","rotatez","z <rot> (degree)",rot_z},
  {"sx","shiftx","sx <shift> (angstrome)",shift_x},
  {"sy","shifty","sy <shift> (angstrome)",shift_y},
  {"sl","select","select <atom> ...",select_atom},
  {"sro","selectrectout","",select_rect_out},
  {"sri","selectrectin","",select_rect_in},
  {"cs","clearstack","cs (no arg)",stack_clr},
  {"rs","restorestack","rs (no arg)",stack_restore},
  {"fr","fragment","fragment <fragment No.>",select_frag},
  {"om","omit","omit atom/cell/label",omit_view},
  {"sc","scale","scale <>",scale_view},
  {"in","include","in atom/cell/label",include_view},
#ifdef WINDOWS
  {"pl","plot","plot ['ps'] ['wmf'] ['color']",plot_ps},
#else
  {"pl","plot","plot ['color']",plot_ps},
#endif
  {"xy","xyz","xy <atom>",dispxyz},
  {"dp","depth","dp <distance (angstrome)>",depth},
  {"s0","shift0","",shift0},
  {"sr","setrot","",rot_step_set},
  {"db","delbond","",del_bond},
  {"mb","makebond","",make_bond},
  {"cb","checkbond","",check_bond},
  {"vw","view","",view},
  {"bo","bond","",bond_list},
  {"rb","resetbond","",reset_bond},
  {"du","dupatom","",dup_atom},
  {"pk","pack","",pack},
  {"pk2","pack2","",pack2},
  {"up","unpack","",unpack},
  {"bu","build","",build_on},
  {"rn","rename","",rename_atom},
  {"na","name","",name},
  {"panic","panic","",panic},
  {"rm","del","",del_atom},
  {"mk","makeatom","",make_atom},
  {"pe","peak","",peak},
  {"or","ortep",
     "ortep [rad [Hrad [SelCode [AnisCode [IsoCode [HCode]]]]]]",ortout},
  {"or3","ortep3",
     "ortep3 [rad [Hrad [SelCode [AnisCode [IsoCode [HCode]]]]]]",ort3out},
  {"cv","cavity","",cavdraw},
  {"pu","purge","",purge},
  {"st","stereo","",stereo},
  {"mo","mono","",mono},
  {"ex","export","",export},
  {"setvmt","setviewmatrix","",setvmt},
  {"cl","color","",setatomcolor},
  {"rc","recover","Recover last session",Recover},
  {"?","shorthelp","? <command>",shorthelp}
};

SV_Function int AppendCommandQueue(char *cmd)
{
  ClearMessage();
  if (nCommandQueue+1>=cBufSize) {
    PutWarning1("Command Queue Over Flow!");
    return ERR;
  }
  strcpy(commandQueue[nCommandQueue++],cmd);
  return OK;
}

SV_Function int InsertCommandQueue(char *cmd)
{
  int i;
  ClearMessage();
  if (nCommandQueue+1>=cBufSize) {
    PutWarning1("Command Queue Over Flow!");
    return ERR;
  }
  nCommandQueue++;
  for(i=nCommandQueue;i>0;i--){
    strncpy(commandQueue[i],commandQueue[i-1],100);
  }
  strcpy(commandQueue[0],cmd);
  return OK;
}

SV_Function int ReadCommandQueue(char *dest)
{
  int i;
  if (nCommandQueue==0) return ERR;
  strcpy(dest,commandQueue[0]);
  nCommandQueue--;
  for(i=0;i<nCommandQueue;i++){
    strcpy(commandQueue[i],commandQueue[i+1]);
  }
  return OK;
}

/* Input Command from Command Queue and Execute */
SV_Function int ExecCommand(void)
{
  int i,j;
  char inst[cBufSize];
  disp_str[0][0]=0;
  if (ReadCommandQueue(fbuf)==ERR) return OK;
#ifdef DEBUG
  PutMessage1(fbuf);
#endif
/*******************
  cmd_c=sscanf(fbuf,"%80s %80s %80s %80s %80s %80s %80s %80s %80s %80s %80s",
                 inst,cmd_param2[0],cmd_param2[1],cmd_param2[2],cmd_param2[3],
                 cmd_param2[4],cmd_param2[5],cmd_param2[6],cmd_param2[7],
                 cmd_param2[8],cmd_param2[9]);
********************/
  cmd_c=0;
  {
    int slen;
    slen=strlen(fbuf);
    for(i=0;i<slen;i++) {
      if (fbuf[i]=='\t') fbuf[i]=' ';
    }
    fbuf[slen]=' ';
    slen++;
    fbuf[slen]=0;
    i=0;
    j=0;
    while(i<slen && fbuf[i]!=' ') {
      inst[j++]=fbuf[i++];
    }
    inst[j]=0;

    /* === ToDo::FixMe:: Rewrite command line decoder */
    if (strcasecmp(inst,"open")==0 || strcasecmp(inst,"openfile")==0) {
      if (i+1<slen) {
	cmd_atom[0]=NULL;
	cmd_num[0]=0;
	cmd_int[0]=0;
	cmd_c=1;
	strcpy(cmd_param[0],fbuf+i+1);
	strcpy(cmd_param2[0],fbuf+i+1);
	SetWaitCursor();
	i=cmd_go(inst);
	UnSetWaitCursor();
	if (i<=0) return i;
	return OK;
      } else {
	return ERR;
      }
    }
    while(i<slen) {
      while(i<slen && fbuf[i]==' ') i++;
      if (fbuf[i]==0) break;
      j=0;
      while(i<slen && fbuf[i]!=' ') {
	cmd_param2[cmd_c][j++]=fbuf[i++];
      }
      cmd_param2[cmd_c++][j]=0;
      i++;
      if (cmd_c>=MAX_CMD_PARAM) break;
    }
  }
/*
  fprintf(stderr,"CMD_C==%d, INST='%s'\n",cmd_c,inst);
  for(i=0;i<cmd_c;i++){
    fprintf(stderr,"CMD[%d]='%s'\n",i,cmd_param2[i]);
  }
*/
/*  if (cmd_c<1) return ERR;*/
  for(i=0;i<cBufSize;i++) {
    if(isalpha(inst[i]))inst[i]=tolower(inst[i]);
    for (j=0;j<cmd_c;j++){
      if(isalpha(cmd_param2[j][i])) {
        cmd_param[j][i]=toupper(cmd_param2[j][i]);
      } else {
        cmd_param[j][i]=cmd_param2[j][i];
      }
    }
  }
  for(i=0;i<cmd_c;i++){
    cmd_atom[i]=getatm(cmd_param[i]);
    cmd_num[i]=atof(cmd_param[i]);
    cmd_int[i]=atoi(cmd_param[i]);
  }
  for (i=cmd_c;i<MAX_CMD_PARAM;i++){
    cmd_atom[i]=NULL;
    cmd_num[i]=0;
    cmd_int[i]=0;
  }
  SetWaitCursor();
  i=cmd_go(inst);
  UnSetWaitCursor();
  if (i<=0) return i;
  return OK;
}

/*
SV_Function void setpara(int a,char *b)
{
  sprintf(cmd_param[a],"%s",b);
  cmd_atom[a]=getatm(b);
  cmd_num[a]=atof(b);
  cmd_int[a]=atoi(b);
}
*/

SV_Function Bool ShiftCommandLine(char *dest)
{
  int i;
  if (cmd_c<1) return ERR;
  if (dest!=NULL)strcpy(dest,cmd_param[0]);
  cmd_c--;
  for(i=0;i<cmd_c-1;i++){
    strcpy(cmd_param[i],cmd_param[i+1]);
    cmd_atom[i]=cmd_atom[i+1];
    cmd_int[i]=cmd_int[i+1];
    cmd_num[i]=cmd_num[i+1];
  }           
  return OK;
}

/* Compare Command/Input */
SV_Function Bool chkinst(char *inst,char *a,char *b)
{
  if (strlen(inst)==0) return FALSE;
  if (strcmp(inst,a)==0) return TRUE;
  if (strcmp(inst,b)==0) return TRUE;
  return FALSE;
}

/* execute command */
SV_LocalFunction int cmd_go(char *inst)
{
  int i;
  char buf[cBufSize];
  if (chkinst(inst,"ex","exit")) return -1;
  if (chkinst(inst,"qu","quit")) return -1;
  if (log_fp != NULL) {
    fprintf(log_fp,"%s\n",fbuf);
    fflush(log_fp);
  }
  if (inst[0]=='#') return 1;
  for(i=0;command[i].func!=NULL;i++) {
    if (chkinst(inst,command[i].cmd1,command[i].cmd2)) {
      command[i].func();
      return 0;
    }
  }
  /* fprintf(stderr,"ERROR: '%s'\n",inst); */
  sprintf(buf,"## unknown command: %s",inst);
  cmderr(buf);
  return 1;
}

SV_Function void shorthelp()
{
  int i;
  char cmd[20];
  if (cmd_c!=1) {
    sprintf(cmd,"?");
  } else {
    for(i=0;i<strlen(cmd_param[0])+1;i++){
      if (isalpha(cmd_param[0][i])) {
	cmd[i]=tolower(cmd_param[0][i]);
      } else {
	cmd[i]=cmd_param[0][i];
      }
    }
  }
  for(i=0;command[i].func!=NULL;i++) {
    if (chkinst(cmd,command[i].cmd1,command[i].cmd2)) {
      cmdinfo(command[i].shortHelp);
      return;
    }
  }
  cmdinfo("ShortHelp Usage:   ? <command>");
}

SV_Function void macro(void)
{
  FILE *cmd_fp;
  if (cmd_c==0) {
    strcpy(fbuf,"sv.mac");
  } else if (cmd_c==1) {
    if (strchr(cmd_param[0],'.')==NULL) {
      sprintf(fbuf,"%s.mac",cmd_param[0]);
    } else {
      strcpy(fbuf,cmd_param[0]);
    }
  }
  /*
  if (cmd_c==1) sprintf(fbuf,"%s%s.cmd",curwd,cmd_param[0]) ;
  else if (cmd_c==0) sprintf(fbuf,"%ssv.cmd",curwd);
  else {
    cmderr("Usage: macro [MacroFileName]");
    return;
  }
  if (cmd_flag!=0) fclose(cmd_fp);
  */
  /* fprintf(stderr,"X%sX\n",fbuf); */
  cmd_fp=ffopen(fbuf,"r");
  if (cmd_fp!=NULL) {
    while(fgets(fbuf,cBufSize,cmd_fp)!=NULL) {
      int len;
      len=strlen(fbuf);
      while(len>0) {
	if (fbuf[len-1]==' ' || fbuf[len-1]=='\n') {
	  fbuf[--len]=0;
	} else {
	  break;
	}
      }
      if (len!=0 && fbuf[0]!='#') AppendCommandQueue(fbuf);
    }
  } else cmderr("Cannot Open Macro File !");
  return;
}

SV_Function void Recover(void)
{
  FILE *cmd_fp;
  if (cmd_c==0) {
    strcpy(fbuf,"sv1.log");
  } else if (cmd_c==1) {
    strcpy(fbuf,cmd_param[0]);
  }
  cmd_fp=ffopen(fbuf,"r");
  if (cmd_fp==NULL) {
    cmderr("Log file not found !");
    return;
  }
    
  fbuf2[0]=0;
  while(fgets(fbuf,cBufSize,cmd_fp)!=NULL) {
    int len;
    len=strlen(fbuf);
    while(len>0) {
      if (fbuf[len-1]==' ' || fbuf[len-1]=='\n' || fbuf[len-1]=='\r') {
	fbuf[--len]=0;
      } else {
	break;
      }
    }
    if (len!=0 && fbuf[0]!='#') {
      if (fbuf2[0]!=0 && fbuf2[0]!='#') {
	AppendCommandQueue(fbuf2);
      }
      strcpy(fbuf2,fbuf);
    }
  } 
  return;
}

SV_Function void plot_ps(void)
{
  int i;

  global_flag |= GLOBALFLAG_PLOT;
#ifndef WINDOWS
    global_flag |= GLOBALFLAG_PSPLOT;
#endif

  for(i=0;i<cmd_c;i++) {
    if (chkinst(cmd_param[i],"COL","COLOR")) {
      global_flag |= GLOBALFLAG_COLORPLOT;
    } 
#ifdef WINDOWS
    if (chkinst(cmd_param[i],"PS","POSTSCRIPT")) {
      global_flag |= GLOBALFLAG_PSPLOT;
    }
    if (chkinst(cmd_param[i],"WMF","EMF")) {
      global_flag |= GLOBALFLAG_WMFPLOT;
    }
#endif
  }
  refresh();
  global_flag &= ~(GLOBALFLAG_PLOT|GLOBALFLAG_COLORPLOT|
		   GLOBALFLAG_PSPLOT|GLOBALFLAG_WMFPLOT);
  output_printer(); /* Unix Only, exec lpr command */
}

SV_LocalFunction void output_printer(void)
{
#ifdef UNIX
  char buf2[200];
  if (getenv("SVPRINT")!=NULL) {
    sprintf(buf2,"%s %s%s",getenv("SVPRINT"),curwd,"sv.ps");
  } else if (getenv("PRINTER")!=NULL) {
    sprintf(buf2,"lpr -P%s %s%s",getenv("PRINTER"),curwd,"sv.ps");
  } else {
    sprintf(buf2,"lpr %s%s",curwd,"sv.ps");
  }
  if (prompt_input("Printout Command :",buf2,buf2,199)==1) {
    system(buf2);
  }
#endif
}

SV_LocalFunction void open_file(void)
{
  if (cmd_c==0) {
    char *FileName;
    FileName=InputFileName("=== Input File Name ===",
                filename(curwd,"shelxl.res"),TRUE);
    if (FileName!=NULL) {
      sprintf(fbuf,"openfile %s",FileName);
      /*      AppendCommandQueue(fbuf);*/
      svx(fbuf);
      refresh();
    } else {
      cmderr("OpenFile Ignored");
    }
  } else if (cmd_c==1) {
    svx(cmd_param2[0]);
  } else {
    cmdinfo("Usage: open [fileName]");
  }
}

SV_LocalFunction void pop_stack(void)
{
  popstack();
}

SV_Function void setvmt(void) {
  int i,j;
  if (cmd_c==1) {
    for(i=0;i<3;i++){
      for(j=0;j<3;j++){
        VMT[i][j]=0;
      }
      VMT[i][i]=1;
    }
    refresh();
  } else if (cmd_c==0) {
    sprintf(fbuf,"setvmt %f %f %f %f %f %f %f %f %f %f %f\n",
            VMT[0][0],VMT[0][1],VMT[0][2],
            VMT[1][0],VMT[1][1],VMT[1][2],
            VMT[2][0],VMT[2][1],VMT[2][2],shift[0],shift[1]);
    cmdinfo(fbuf);
  } else if (cmd_c==9 || cmd_c==11) {
    for(i=0;i<3;i++){
      for(j=0;j<3;j++){
        VMT[i][j]=cmd_num[i*3+j];
      }
    }
    if (cmd_c==11) {
      shift[0]=cmd_num[9];
      shift[1]=cmd_num[10];
    }
    refresh();
  } else {
    cmdinfo("Usage: "
           "setvmt [a11 a12 a13 a21 a22 a23 a31 a32 a33 [sx sy]] "
	   "  or  setvmt 0");
  }
}

SV_Function void select_frag(void)
{
  int ffrag,i,j;
  for(i=0;i<n_satom;i++) SAT[i].view_flag|=FLAG_OMIT;
  for(j=0;j<cmd_c;j++){
    ffrag=cmd_int[j];
    for (i=0;i<n_satom;i++) {
      if (ffrag==0||SAT[i].frag==ffrag 
        || SAT[i].frag==-ffrag || (SAT[i].frag==0 && ffrag<0)) {
        SAT[i].view_flag&= ~FLAG_OMIT;
      }
    }
  }
  refresh();
}
        
SV_Function void future(void)
{
/*  cmderr("This command will be available in the future.");*/
  cmderr("Reserved Command");
}

SV_LocalFunction void refresh_sub(void (*GCOLOR)(int),void (*GLINE)(Double,Double,Double,Double),void (*GPUTS)(Double,Double,char *))
{
  int plot_flag;
  if ((global_flag & GLOBALFLAG_STEREO)!=0) {
    Double tmpvmt[3][3];
    int i,j;
    for(i=0;i<3;i++){
      for(j=0;j<3;j++){
        tmpvmt[i][j]=VMT[i][j];
      }
    }
    roty(-1.5*PI/180);
    sshift[0]=-ssplit;
    calc_atoms();
    shelview(GCOLOR,GLINE,GPUTS);

    sshift[0]=ssplit;
    roty(3.0*PI/180);
    calc_atoms();

    shelview(GCOLOR,GLINE,GPUTS);

    sshift[0]=0;
    for(i=0;i<3;i++){
      for(j=0;j<3;j++){
        VMT[i][j]=tmpvmt[i][j];
      }
    }
  } else {
    calc_atoms();
    shelview(GCOLOR,GLINE,GPUTS);
  }
}

SV_Function void refresh(void)
{
  int plot_flag,ps_flag,wmf_flag,color_flag;
  if ((global_flag & GLOBALFLAG_NOTREFRESH)!=0) { return; }
  if (n_satom==0) return;

  plot_flag=(global_flag & GLOBALFLAG_PLOT)!=0;
  ps_flag=(global_flag & GLOBALFLAG_PSPLOT)!=0;
  wmf_flag=(global_flag & GLOBALFLAG_WMFPLOT)!=0;
  color_flag=(global_flag & GLOBALFLAG_WMFPLOT)!=0;
  
  if (plot_flag) {
    global_flag &= ~GLOBALFLAG_PLOT;
    if (ps_flag) {
      if (PSopen()==0) {
	PutMessage1("PS File Open Error\n");
      } else {
	refresh_sub(PScolor,PSline,PSputs);
	PSclose(); 
      }
    }
#ifdef WINDOWS
    if (wmf_flag) {
      if (EMFopen()==0) {
	PutMessage1("WMF File Open Error\n");
      } else {
	refresh_sub(gcolor,gline,gputs);
	EMFclose(); 
      }
    }
    if (!ps_flag && !wmf_flag) {
      if (LPTopen()==ERR) {
	/*
	PutMessage1("Error: Printer Initialize\n");
	*/
      } else {
	refresh_sub(gcolor,gline,gputs);
	LPTclose();
      }
      return;
    }
#endif
  } 
  gopen();
  gcls();
  refresh_sub(gcolor,gline,gputs);
  if ((global_flag & GLOBALFLAG_TMP_LABEL)!=0) {
    putstr();
    disp_atom_stack();
  }
  gclose();
}

SV_Function void stereo(void)
{
  if ((global_flag & GLOBALFLAG_STEREO)==0) vscale*=0.5;
  global_flag|=GLOBALFLAG_STEREO;
  if (cmd_c==0 || cmd_num[0]==0) ssplit=2.5; else ssplit=cmd_num[0];
  refresh();
  return;
}

SV_Function void mono(void)
{
  if (global_flag & GLOBALFLAG_STEREO) vscale*=2;
  global_flag&=~GLOBALFLAG_STEREO;
  refresh();
  return;
}
