/* File Export Module   "file.c"     -*- C -*-  */

/*
#S#
#
# This Program is part of SV (Shelx-output Viewer)
# Copyright (C) 1991-1998 Takashi Nemoto
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details. (File "COPYING")
#
#    Note!: General Public Licence is copyrighted by the Free Software 
#           Foundation,  but this code is copyrighted by me.
#
#    Send bugs and comments to tnemoto@eels.kuicr.kyoto-u.ac.jp
#
#E#
*/

#define __file_c__
#define __main_c_proc__
#define __version_c_proc__
#define __message_c_proc__
#define __readf_c_proc__
#define __shelplot_c_proc__
#define __shelview_c_proc__
#define __cmd_c_proc__
#define __atom_c_proc__
#define __bond_c_proc__
#define __subs_c_proc__

#define __ortep_c_proc__
#define __cavity_c_proc__
#define __charon_c_proc__
#define __povray_c_proc__


#define __string_proc__

#define __syssub__

#include "def.h"
#include <math.h>
#include <ctype.h>

SV_LocalFunction const char* float2frac(Double f,Bool printOne)
{
  const char* fracstr[]={ "0","1/12","1/6","1/4","1/3","5/12","1/2",
			  "7/12","2/3","3/4","5/6","11/12","1"};
  int sign,i;
  static char buf[16];
  Double f2;
  sign=f>=0?1:0;
  f2=fabs(f);
  for(i=0;i<13;i++){
    if (fabs((Double)i/12.0-f2)<0.0001){
      if (i==12 && !printOne) {
	sprintf(buf,"%c",sign?'+':'-');
      } else {
	sprintf(buf,"%c%s",sign?'+':'-',fracstr[i]);
      }
      return buf;
    }
  }
  sprintf(buf,"%+f",f);
  return buf;
}

SV_LocalFunction const char* symm2str(Double sym[4])
{
  static char symstr[80];
  char buf[80];
  int i;
  symstr[0]=0;
  for(i=0;i<3;i++){
    if (sym[i+1]!=0.0){
      sprintf(buf,"%s%c",float2frac(sym[i+1],FALSE),"xyz"[i]);
      strcat(symstr,buf);
    }
  }
  if (sym[0]!=0){
    strcat(symstr,float2frac(sym[0],TRUE));
  }
  return symstr;
}

SV_Function void export(void) 
{
  char exportType[cBufSize];
  if (cmd_c<1) {
    cmderr("Usage: export <FileType>");
    return;
  }
  ShiftCommandLine(exportType);
  if (chkinst(cmd_param[0],"SHELXL","SH")) {
    shelxout();
  } else if (chkinst(cmd_param[0],"ORTEP","OR")) {
    ortout();
  } else if (chkinst(cmd_param[0],"CAVITY","CV")) {
    cavdraw();
  } else if (chkinst(cmd_param[0],"CHARON","CH")) {
    charonout();
  } else if (chkinst(cmd_param[0],"POVRAY","PV")) {
    povrayout();
  } else if (chkinst(cmd_param[0],"CHEMRAY","CR")) {
    chemrayout();
  } else if (chkinst(cmd_param[0],"CHEMDRAW","CD")) {
    chemdrawout();
  } else if (chkinst(cmd_param[0],"RASMOL","RA")) {
    rasmolout();
  } else if (chkinst(cmd_param[0],"CSD","FDAT")) {
    fdatout();
  } else if (chkinst(cmd_param[0],"MACTEMPAS","MACTEMPASS")) {
    mactempasout();
  } else {
    cmderr("export: unknow file type");
  }
}

SV_Function void shelxout(void)
{
  FILE *fp;
  int i;
  Double *loc;
  Double countAtom[MAX_SFAC];

  for(i=0;i<n_sfac;i++) countAtom[i]=0;
  for(i=0;i<n_atom;i++) {
    countAtom[orgAtom[i].atom_type-1]+=orgAtom[i].occ;
  }
  fp=ffopen(InputFileName("Export File (Shelxl Type)","svx.res",0),"w");
  if (fp==NULL) {
    cmderr("Can't open output file!");
    return;
  }
  fprintf(fp,"TITL %s\n",Title);
  fprintf(fp,"CELL %10.5f%10.5f%10.5f%10.5f%10.5f%10.5f%10.5f\n",
          1.54184,cell[0],cell[1],cell[2],cell[3]*180/PI,cell[4]*180/PI,
        cell[5]*180/PI);
  fprintf(fp,"ZERR %10d%10.5f%10.5f%10.5f%10.5f%10.5f%10.5f\n",
          symnum,0.01,0.01,0.01,0.01,0.01,0.01);
  fprintf(fp,"LATT %d\n",-1);
/*  fprintf(fp,"SYMM %d\n",-1);*/
  fprintf(fp,"SFAC ");
  for(i=0;i<n_sfac;i++) fprintf(fp,"%-2s ",sfac[i].name);
  fprintf(fp,"\n");
  fprintf(fp,"UNIT ");
  for(i=0;i<n_sfac;i++) fprintf(fp,"%7.2f ",countAtom[i]);
  fprintf(fp,"\n");
  fprintf(fp,"LIST -1\n");
  fprintf(fp,"FMAP -2\n");
  fprintf(fp,"PLAN 50\n");
  fprintf(fp,"TEMP 23\n");
  fprintf(fp,"L.S. 3\n");
  for(i=0;i<n_satom;i++){
    if (OrgAtom(i)->name[0]=='Q') continue;
    loc=sym_exp(SAT+i,orgAtom);
    fprintf(fp,"%-4s %2d %10.6f %10.6f %10.6f %10.6f %10.6f\n",
            OrgAtom(i)->name,OrgAtom(i)->atom_type,
            loc[0],loc[1],loc[2],
            11.0,0.05);
  }
  fprintf(fp,"HKLF 4\n");
  fclose(fp);
}

SV_LocalFunction int EnumAtoms(void) 
{
  int i,n_atom2=0;
  for(i=0;i<n_atom;i++){
    orgAtom[i].temp_num=0;
  }
  for(i=0;i<n_satom;i++){
    if ((SAT[i].view_flag & FLAG_OMIT)==0
        && OrgAtom(i)->name[0]!='Q' && OrgAtom(i)->temp_num==0) {
      OrgAtom(i)->temp_num=++n_atom2;
    }
  }
  return n_atom2;
}

SV_LocalFunction int EnumSAT(void) 
{
  int i,n_atom2=0;
  for(i=0;i<n_satom;i++){
    SAT[i].temp_num=0;
  }
  for(i=0;i<n_satom;i++){
    if ((SAT[i].view_flag & FLAG_OMIT)==0
        && OrgAtom(i)->name[0]!='Q' && OrgAtom(i)->temp_num==0) {
      SAT[i].temp_num=++n_atom2;
    }
  }
  return n_atom2;
}

SV_Function void fdatout(void)
{
  FILE *fp;
  int i,j;
  int *B1,*B2,*A1,n_bond2,n_bond3,n_atom2;
  char FormatString[4];
  int n_param,max_param;

  B1=(int *)Malloc(sizeof(int)*n_bond+1);
  if (B1==NULL) {
    cmderr("Out of Memory!");
    return;
  }
  B2=(int *)Malloc(sizeof(int)*n_bond+1);
  if (B2==NULL) {
    Free(B1);
    cmderr("Out of Memory!");
    return;
  }
  fp=ffopen(InputFileName("Export File (CSD FDAT Type)","svx.dat",0),"w");
  if (fp==NULL) {
    cmderr("Can't open output file!");
    Free(B2);
    Free(B1);
    return; 
  };

  n_bond2=0;
  n_bond3=0;
  n_atom2=EnumAtoms();
  for(i=0;i<n_bond;i++){
    int *a1,*a2;
    int b1,b2;
    int sameflag=0;
    a1=&(OrgAtom(BOND[i].atm[0])->temp_num);
    a2=&(OrgAtom(BOND[i].atm[1])->temp_num);
    b1=abs(*a1);
    b2=abs(*a2);
    if (b1==0 || b2==0) continue;
    if (*a1<0 && *a2<0){
      for(j=0;j<n_bond2;j++){
        if ((*a1==B1[j] && *a2==B2[j]) || (*a1==B2[j] && *a2==B1[j])) {
          sameflag=1;
          break;
        }
      }
    }
    if (sameflag==0) {
      B1[n_bond2]=b1;
      B2[n_bond2++]=b2;
      *a1=-b1;
      *a2=-b2;
    }
  }

  /* recover Atom Number */
  for(i=0;i<n_atom;i++){
    orgAtom[i].temp_num=abs(orgAtom[i].temp_num);
  }

  A1=(int *)Malloc(sizeof(int)*n_atom2+1);
  if (A1==NULL) {
    Free(B2);
    Free(B1);
    fclose(fp);
    cmderr("Out of Memory!");
    return;
  }

  for(i=0;i<n_atom2;i++){
    A1[i]=0;
  }
  for(i=0;i<n_bond2;i++){
    if (A1[B1[i]-1]==0) {
      A1[B1[i]-1]=B2[i];
      B1[i]=0;
      n_bond3++;
    } else if (A1[B2[i]-1]==0) {
      A1[B2[i]-1]=B1[i];
      B1[i]=0;
      n_bond3++;
    }
  }
  fprintf(fp,
/*  0000000001111111111222222222233333333334444444444555555555566666666667 */
/*  1234567890123456789012345678901234567890123456789012345678901234567890 */
   "#SVXDUMMY03111111      %3d  0  0  0  0%3d%3d%3d%3d   %3d1321001"
          "0000000000000011\n",
          100,symnum,n_sfac,n_atom2,0,
          (n_bond2-n_bond3)*2>999?998:(n_bond2-n_bond3)*2);
  fprintf(fp,"%6d%6d%6d%6d%6d%6d333333000000000000000000             0 0\n",
  (int)(cell[0]*1000),(int)(cell[1]*1000),(int)(cell[2]*1000),
  (int)(cell[3]*180000/PI),(int)(cell[4]*180000/PI),(int)(cell[5]*180000/PI));
/*  fprintf(fp,"\n");*/
  n_param=0;
  for(i=0;i<symnum;i++){
    fprintf(fp,"%1d%1d%1d%2d%1d%1d%1d%2d%1d%1d%1d%2d",
            (int)symm[i][0][1]+1,(int)symm[i][0][2]+1,(int)symm[i][0][3]+1,
            (int)symm[i][0][0]*12,
            (int)symm[i][1][1]+1,(int)symm[i][1][2]+1,(int)symm[i][1][3]+1,
            (int)symm[i][1][0]*12,
            (int)symm[i][2][1]+1,(int)symm[i][2][2]+1,(int)symm[i][2][3]+1,
            (int)symm[i][2][0]*12
            );
    n_param++;
    if (n_param==5){ n_param=0; fprintf(fp,"     \n"); }
  }
  if (n_param!=0) { n_param=0; fprintf(fp,"     \n"); }
  n_param=0;
  for(i=0;i<n_sfac;i++){ 
    fprintf(fp,"%-2s%3d",sfac[i].name,(int)(sfac[i].vdw*100));
  };
  fprintf(fp,"\n"); 
  for(i=0;i<n_atom;i++){
    if (orgAtom[i].temp_num==0) continue;
    fprintf(fp,"%-5s%7d%7d%7d",
            orgAtom[i].name,
            (int)(100000*orgAtom[i].loc[0]),
            (int)(100000*orgAtom[i].loc[1]),
            (int)(100000*orgAtom[i].loc[2]));
    n_param++;
    if(n_param==3){
      n_param=0; fprintf(fp,"\n"); 
    } else { 
      fprintf(fp," ");
    }
  }
  if(n_param!=0){fprintf(fp,"\n"); };
  if (n_atom2<100) {
    max_param=40;
    strcpy(FormatString,"%2d");
  } else {
    max_param=26;
    strcpy(FormatString,"%3d");
  }
  n_param=0;
  for(i=0;i<n_atom2;i++){
    fprintf(fp,FormatString,A1[i]);
    n_param++;
    if (n_param==max_param) {
      n_param=0;
      fprintf(fp,"\n");
    }
  }
  for(i=0;i<n_bond2;i++){
    if (B1[i]==0) continue;
    fprintf(fp,FormatString,B1[i]); 
    n_param++;
    if (n_param==max_param) {
      n_param=0;
      fprintf(fp,"\n");
    }
    fprintf(fp,FormatString,B2[i]); 
    n_param++;
    if (n_param==max_param) {
      n_param=0;
      fprintf(fp,"\n");
    }
  }
  if (n_param!=0) fprintf(fp,"\n");
  Free(A1);
  Free(B2);
  Free(B1);
  fclose(fp);
}

SV_Function void chemdrawout(void) 
{
  FILE *fp;
  int n_atom2,n_bond2,i;
  char ElementName[3];
  ElementName[2]=0;
  fp=ffopen(InputFileName("Export (ChemDraw Connection Table Type)","svx.ct",0),"w");
  if (fp==NULL) {
    cmderr("Can't open output file!");
    return;
  }
  n_atom2=EnumSAT();
  n_bond2=0;
  for(i=0;i<n_bond;i++){
    if (SAT[BOND[i].atm[0]].temp_num!=0 && 
        SAT[BOND[i].atm[1]].temp_num!=0 &&
	SAT[BOND[i].atm[0]].temp_num<SAT[BOND[i].atm[1]].temp_num
	) {
      n_bond2++;
    }
  }
  fprintf(fp,"%s\n",Title);
  fprintf(fp,"%d %d\n",n_atom2,n_bond2);
  for(i=0;i<n_satom;i++){
    if (SAT[i].temp_num!=0) {
      strcpy(ElementName,sfac[OrgAtom(i)->atom_type-1].name);
      if (isalpha(ElementName[1])) ElementName[1]=tolower(ElementName[1]);
      fprintf(fp,"%10.4f %10.4f %10.4f %2s\n",
              SAT[i].axy[0],SAT[i].axy[1],SAT[i].axy[2],
              ElementName);
    }
  }
  for(i=0;i<n_bond;i++){
    if (SAT[BOND[i].atm[0]].temp_num!=0 && 
        SAT[BOND[i].atm[1]].temp_num!=0 &&
	SAT[BOND[i].atm[0]].temp_num<SAT[BOND[i].atm[1]].temp_num
	) {
      fprintf(fp,"%d %d 1 1\n",SAT[BOND[i].atm[0]].temp_num,
              SAT[BOND[i].atm[1]].temp_num);
    }
  }
  fclose(fp);
}

SV_Function void mactempasout(void)
{
  FILE *fp;
  int i;
  char ElementName[3];
  ElementName[2]=0;
  fp=ffopen(InputFileName("Export (MacTempas)","svx.at",0),"w");
  if (fp==NULL) {
    cmderr("Can't open output file!");
    return;
  }
  fprintf(fp,"%s\n",Title);
  fprintf(fp,"0 ### Space Group Number\n");
  fprintf(fp,"%f %f %f %f %f %f\n",
	  cell[0],cell[1],cell[2],
	  cell[3]/PI*180.0,cell[4]/PI*180,cell[5]/PI*180);
  fprintf(fp,"1.0\n");
  fprintf(fp,"%f %f %f\n",0.0,0.0,1.0);
  fprintf(fp,"%d 1\n",symnum);
  fprintf(fp,"%d %d\n",n_atom,n_sfac);
  for(i=0;i<n_atom;i++){
      strcpy(ElementName,sfac[orgAtom[i].atom_type-1].name);
      if (isalpha(ElementName[1])) ElementName[1]=tolower(ElementName[1]);
      fprintf(fp,"%d %s %f %f %f %f %f\n",
	      orgAtom[i].atom_type,ElementName,
	      orgAtom[i].loc[0],orgAtom[i].loc[1],orgAtom[i].loc[2],
	      get_ueq(orgAtom+i),orgAtom[i].occ);
  }
  fprintf(fp,"200CX 1.2 100 0.65\n");
  fprintf(fp,"200.00\n");
  fprintf(fp,"0.0 0.0\n");
  fprintf(fp,"50.0,100.0,10.0\n");
  fprintf(fp,"NO\n");
  fprintf(fp,"-600,0,100\n");
  fprintf(fp,"0.7\n");
  fprintf(fp,"0 0 0 0\n");
  for(i=0;i<symnum;i++){
    fprintf(fp,"%s,",symm2str(symm[i][0]));
    fprintf(fp,"%s,",symm2str(symm[i][1]));
    fprintf(fp,"%s\n",symm2str(symm[i][2]));
  }
  fprintf(fp,"1\n");
  fprintf(fp,"0\n");
   
  fclose(fp);
}
