/* input external coordinates "readmeta.c"  -*- C -*- */

/*
#S#
#
# This Program is part of SV (Shelx-output Viewer)
# Copyright (C) 1991-1998 Takashi Nemoto
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details. (File "COPYING")
#
#    Note!: General Public Licence is copyrighted by the Free Software 
#           Foundation,  but this code is copyrighted by me.
#
#    Send bugs and comments to tnemoto@eels.kuicr.kyoto-u.ac.jp
#
#E#
*/

#define __readmeta_c__
#define __readf_c_proc__
#define __cmd_c_proc__
#define __subs_c_proc__
#define __string_proc__
#include <math.h>

#include "def.h"

#ifdef IGraph

SV_Function void readmeta(void)
{
  FILE *meta_fp;
  IGraph_Lines *ili;
  IGraph_Arcs *iar;
  int i,j,col;
  char dummy[10];
  global_flag&=~GLOBALFLAG_IGRAPH;
  if (cmd_c==0) {
    cmderr("IGraph Off");
    refresh();
    return;
  }
  if (cmd_c!=1) {
    cmdinfo("Usage: readmeta <FileName>");
    return;
  }
  if ((meta_fp=ffopen(cmd_param2[0],"r"))==NULL) {
    cmderr("ReadMeta: Can Not Open File");
    return;
  }
  do {
    if (ffgets(meta_fp)==0) {
      f_err(meta_fp);
      return;
    };
  } while (strcmp("%# StartData",fbuf)!=0);
  n_iarcs=0;
  n_ilines=0;
  n_iorgs=1;
  n_irvmts=1;
  for(i=0;i<3;i++){
    for(j=0;j<3;j++){
      irvmts[0][i][j]=0;
    }
    irvmts[0][i][i]=1;
    iorgs[0][i]=0;
  };
  while(ffgets(meta_fp)!=0 && strcmp("%# EndData",fbuf)!=0) {
    if (strncmp("%# RVMT",fbuf,7)==0) {
      sscanf(fbuf+10,"%lf %lf %lf",
	     irvmts[n_irvmts][0],irvmts[n_irvmts][0]+1,irvmts[n_irvmts][0]+2);
      ffgets(meta_fp);
      sscanf(fbuf+10,"%lf %lf %lf",
	     irvmts[n_irvmts][1],irvmts[n_irvmts][1]+1,irvmts[n_irvmts][1]+2);
      ffgets(meta_fp);
      sscanf(fbuf+10,"%lf %lf %lf",
	     irvmts[n_irvmts][2],irvmts[n_irvmts][2]+1,irvmts[n_irvmts][2]+2);
      n_irvmts++;
    }
    if (strncmp("%# ORIGIN=",fbuf,10)==0) {
      sscanf(fbuf+10,"%lf %lf %lf",
	     iorgs[n_iorgs],iorgs[n_iorgs]+1,iorgs[n_iorgs]+2);
      n_iorgs++;
    }
    if (fbuf[0]!='c') continue;  /* Not Command */
    switch(fbuf[1]) {
    case 'l':
      if (n_ilines>=MAX_ILINES) break;
      ili=ilines+n_ilines;
      i=sscanf(fbuf,"%3s %lf %lf %lf %lf %1s %1s %lf %lf",dummy,
	       ili->vloc,ili->vloc+1,ili->vloc2,ili->vloc2+1,
	       dummy,dummy,ili->vloc+2,ili->vloc2+2);
      ili->col=col;
      ili->matnum=n_irvmts-1;
      ili->orgnum=n_iorgs-1;
      if (i==9) n_ilines++;
      break;
    case 'a':
      if (n_iarcs>=MAX_IARCS) break;
      iar=iarcs+n_iarcs;
      i=sscanf(fbuf,"%3s %lf %lf %lf %lf %lf %1s %1s %lf",dummy,
	       iar->vloc,iar->vloc+1,&(iar->r),&(iar->st),&(iar->ed),
	       dummy,dummy,iar->vloc+2);
      iar->col=col;
      iar->matnum=n_irvmts-1;
      iar->orgnum=n_iorgs-1;
      if (i==9) n_iarcs++;
      break;
    case 'p':
      i=sscanf(fbuf,"%3s %d",dummy,&col);
      break;
    default:
      break;
    }
  }
    fclose(meta_fp);
  global_flag|=GLOBALFLAG_IGRAPH;
  refresh();
}
#else
SV_Function void readmeta(void)
{
  cmderr("This command is disabled.");
}
#endif
