/* Cmd SubModule  "bond.c"           -*- C -*- */

/*
#S#
#
# This Program is part of SV (Shelx-output Viewer)
# Copyright (C) 1991-1998 Takashi Nemoto
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details. (File "COPYING")
#
#    Note!: General Public Licence is copyrighted by the Free Software 
#           Foundation,  but this code is copyrighted by me.
#
#    Send bugs and comments to tnemoto@eels.kuicr.kyoto-u.ac.jp
#
#E#
*/

#define __stack_c__
#define __cmd_c_proc__
#define __subs_c_proc__
#define __atom_c_proc__
#define __shelplot_c_proc__
#ifdef Xt
#define __xtcalls_c_proc__
#endif

#include <math.h>
#include <ctype.h>
#include <string.h>
#include "def.h"

SV_Variable int *atom_stack,*atom_stack_bak;
SV_Variable int atom_stack_top,atom_stack_top_bak;
SV_Variable int MAX_ATOM_STACK,MAX_ATOM_STACK_BAK;

SV_Function void stack_clr(void)
{
  int i;
  for(i=0;i<=atom_stack_top;i++) {
    SAT[atom_stack[i]].view_flag&=~FLAG_SEL;
  };
  if (atom_stack_bak!=0) Free(atom_stack_bak);
  atom_stack_bak=atom_stack;
  atom_stack_top_bak=atom_stack_top;
  MAX_ATOM_STACK_BAK=MAX_ATOM_STACK;
  atom_stack_top=-1;
  atom_stack=NULL;
  MAX_ATOM_STACK=-1;
  refresh();
}

SV_Function void stack_restore(void)
{
  int i;
  S_AT *at;
  int *X;
  for(i=0;i<=atom_stack_top;i++) {
    at=SAT+atom_stack[i];
    if(at!=NULL) at->view_flag&=~FLAG_SEL;
  };
  X=atom_stack;
  atom_stack=atom_stack_bak;
  atom_stack_bak=X;
  i=MAX_ATOM_STACK;
  MAX_ATOM_STACK=MAX_ATOM_STACK_BAK;
  MAX_ATOM_STACK_BAK=i;
  i=atom_stack_top;
  atom_stack_top=atom_stack_top_bak;
  atom_stack_top_bak=i;
  for(i=0;i<=atom_stack_top;i++) {
    at=SAT+atom_stack[i];
    if (at!=NULL) at->view_flag|=FLAG_SEL;
  };
  refresh();
}

SV_Function void AppendAtomToStack(int num)
{
  atom_stack_top++;
  if (atom_stack_top>=MAX_ATOM_STACK) {
    atom_stack=Realloc(atom_stack,&MAX_ATOM_STACK,
		       STACK_BLOCK_SIZE,sizeof(int));
    if (atom_stack==NULL) {
      PutMessage1("Out of Memory");
      Exit(1);
    }
  }
  atom_stack[atom_stack_top]=num;
}

SV_Function void BackupStack(void)
{
  if (atom_stack_bak!=NULL) Free(atom_stack_bak);
  atom_stack_bak=Malloc(MAX_ATOM_STACK*sizeof(int));
  if (atom_stack_bak==NULL){
    PutMessage1("Out of Memory");
    Exit(1);
  }
  memcpy(atom_stack_bak,atom_stack,MAX_ATOM_STACK*sizeof(int));
  atom_stack_top_bak=atom_stack_top;
  MAX_ATOM_STACK_BAK=MAX_ATOM_STACK;
  refresh();
}

SV_Function void select_atom(void)
{
  int i,j,p=0;
  S_AT *at;
  for(j=0;j<cmd_c;j++){
    if (chkinst(cmd_param[j],"PA","PART")){
      j++;
      if (cmd_atom[j]!=NULL) {
	p=cmd_atom[j]->part;
      } else p=cmd_int[j];
      for(i=0;i<n_satom;i++){
	if (((SAT[i].part==p || SAT[i].part==0) && p>=0) ||  SAT[i].part==-p
	  ) {
          SAT[i].view_flag|=FLAG_SEL;
          AppendAtomToStack(i);
        }
      }
    } else if (chkinst(cmd_param[j],"FR","FRAGMENT")){
      j++;
      if (cmd_atom[j]!=NULL) {
	p=cmd_atom[j]->frag;
      } else p=cmd_int[j];
      for(i=0;i<n_satom;i++){
	if (((SAT[i].frag==p || SAT[i].frag==0) && p>=0) ||  SAT[i].frag==-p
	  ) {
          SAT[i].view_flag|=FLAG_SEL;
          AppendAtomToStack(i);
        }
      }
    } else if (chkinst(cmd_param[j],"INV","INVERT") || 
	       chkinst(cmd_param[j],"INVERSE","INV")){
      atom_stack_top=-1;
      for(i=0;i<n_satom;i++){
	if (SAT[i].view_flag & FLAG_SEL) {
	  SAT[i].view_flag &= ~FLAG_SEL;
	} else {
          SAT[i].view_flag|=FLAG_SEL;
          AppendAtomToStack(i);
        }
      }
    } else if (cmd_param[j][0]=='$') {
      int k,f;
      f=-1;
      if (cmd_param[j][2]==0){
	cmd_param[j][2]=' ';
	cmd_param[j][3]=0;
      }
      for(k=0;k<n_sfac;k++){
	if (strcasecmp(sfac[k].name,cmd_param[j]+1)==0) {
	  f=k;
	  break;
	}
      }
      if (f==-1) {
	cmderr("Element Not Found !");
	return;
      }
      f++;
      for(k=0;k<n_satom;k++){
	if (OrgAtom(k)->atom_type==f){
	  SAT[k].view_flag|=FLAG_SEL;
	  AppendAtomToStack(SAT[k].num);
	}
      }
    } else {
      at=cmd_atom[j];
      if (at==NULL) continue ;
      at->view_flag|=FLAG_SEL;
      AppendAtomToStack(at->num);
    }
    dispxyz();
  }
}

SV_Function void select_atom1(S_AT *at)
{
  at->view_flag|=FLAG_SEL;
  AppendAtomToStack(at->num);
}

SV_Function S_AT *popstack(void) 
{
  int i;
  if (atom_stack_top<0) {
    cmderr("Empty Stack !!");
    refresh();
    return NULL;
  } else {
    stacktop->view_flag&=~FLAG_SEL;
    for(i=0;i<atom_stack_top;i++) {
       SAT[atom_stack[i]].view_flag|=FLAG_SEL;
    }
    refresh();
    return SAT+atom_stack[atom_stack_top--];
  }
}


SV_Function void disp_atom_stack(void) 
{
  int i;
  for(i=0;i<10 && i<=atom_stack_top;i++) 
    atom_stack_s[i]=SAT+atom_stack[atom_stack_top-i];
  for(i=0;i<9 && i<atom_stack_top;i++)   
    atom_st_bnd[i]=get_length(atom_stack_s[i],atom_stack_s[i+1]);
  for(i=0;i<8 && i<atom_stack_top-1;i++)
    atom_st_ang[i]=get_angle(atom_stack_s[i],
			     atom_stack_s[i+1],atom_stack_s[i+2]);
  for(i=0;i<7 && i<atom_stack_top-2;i++)  
    atom_st_tor[i]=get_torsion(atom_stack_s[i],atom_stack_s[i+1],
			       atom_stack_s[i+2],atom_stack_s[i+3]);
  gdisp_atom_stack();
}

SV_LocalVariable Double select_rectmin[2],select_rectmax[2];
SV_LocalFunction Bool select_rect(void)
{
  Double t;
  int i,c=0;
  if (cmd_atom[0]!=NULL) {
    for(i=0;i<2;i++) select_rectmin[i]=cmd_atom[c]->axy[i];
    c++;
  } else {
    select_rectmin[0]=cmd_num[c++];
    select_rectmin[1]=cmd_num[c++];
  };
  if (cmd_atom[c]!=NULL) {
    for(i=0;i<2;i++) select_rectmax[i]=cmd_atom[c]->axy[i];
    c++;
  } else {
    select_rectmax[0]=cmd_num[c++];
    select_rectmax[1]=cmd_num[c++];
  };
  if (c!=cmd_c) {
    cmderr("Select_Rect: Parameter Error!");
    return ERR;
  }
  for(i=0;i<2;i++){
    if (select_rectmin[i]>select_rectmax[i]) {
      t=select_rectmin[i];
      select_rectmin[i]=select_rectmax[i];
      select_rectmax[i]=t;
    }
  }
  return OK;
}

SV_Function void select_rect_in(void)
{
  int i;
  if (select_rect()==ERR) return;
  for(i=0;i<n_satom;i++){
    if (SAT[i].axy[0]<select_rectmax[0] && SAT[i].axy[0]>select_rectmin[0] &&
	SAT[i].axy[1]<select_rectmax[1] && SAT[i].axy[1]>select_rectmin[1] &&
	(SAT[i].view_flag & FLAG_OMIT)==0) {
      select_atom1(SAT+i);
    }
  }
  refresh();
}

SV_Function void select_rect_out(void)
{
  int i;
  if (select_rect()==ERR) return;
  for(i=0;i<n_satom;i++){
    if ((SAT[i].axy[0]>select_rectmax[0] || SAT[i].axy[0]<select_rectmin[0] ||
	SAT[i].axy[1]>select_rectmax[1] || SAT[i].axy[1]<select_rectmin[1]) &&
	(SAT[i].view_flag & FLAG_OMIT)==0) {
      select_atom1(SAT+i);
    }
  }
  refresh();
}


