/* Atom Setup/Information  "subs.c"           -*- C -*- */

/*
#S#
#
# This Program is part of SV (Shelx-output Viewer)
# Copyright (C) 1991-1998 Takashi Nemoto
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details. (File "COPYING")
#
#    Note!: General Public Licence is copyrighted by the Free Software 
#           Foundation,  but this code is copyrighted by me.
#
#    Send bugs and comments to tnemoto@eels.kuicr.kyoto-u.ac.jp
#
#E#
*/

#define __subs_c__
#define __shelplot_c_proc__
#define __syssub__

#include <math.h>
#include <ctype.h>
#include "def.h"

SV_Function void inv_matrix(Double org[3][3],Double dest[3][3])
{
  Double norm;
  norm=org[0][0]*org[1][1]*org[2][2]+org[1][0]*org[2][1]*org[0][2]
      +org[2][0]*org[0][1]*org[1][2]-org[0][0]*org[2][1]*org[1][2]
      -org[1][0]*org[0][1]*org[2][2]-org[2][0]*org[1][1]*org[0][2];
  dest[0][0]=1/norm*(org[1][1]*org[2][2]-org[2][1]*org[1][2]);
  dest[0][1]=-1/norm*(org[0][1]*org[2][2]-org[2][1]*org[0][2]);
  dest[0][2]=1/norm*(org[0][1]*org[1][2]-org[1][1]*org[0][2]);
  dest[1][0]=-1/norm*(org[1][0]*org[2][2]-org[2][0]*org[1][2]);
  dest[1][1]=1/norm*(org[0][0]*org[2][2]-org[2][0]*org[0][2]);
  dest[1][2]=-1/norm*(org[0][0]*org[1][2]-org[1][0]*org[0][2]);
  dest[2][0]=1/norm*(org[1][0]*org[2][1]-org[2][0]*org[1][1]);
  dest[2][1]=-1/norm*(org[0][0]*org[2][1]-org[2][0]*org[0][1]);
  dest[2][2]=1/norm*(org[0][0]*org[1][1]-org[1][0]*org[0][1]);
}

SV_Function Long EncodeAdc(S_AT *at)
{
  return ((Long)55500)+at->sym+at->sft[0]*10000+at->sft[1]*1000
           +at->sft[2]*100;
}

SV_Function S_AT *DecodeAdc(Long adc)
{
  static S_AT at;
  int i;
  i=adc/100000;
  at.atm=i-1;
  adc-=i;
  at.sft[0]=adc/10000-5;
  adc-=at.sft[0]*10000+50000;
  at.sft[1]=adc/1000-5;
  adc-=at.sft[1]*1000+5000;
  at.sft[2]=adc/100-5;
  adc-=at.sft[2]*100+500;
  at.sym=adc;
  return &at;
}

SV_Function Bool check_newatom(S_AT *at)
{
  int i;
  for(i=0;i<n_atom;i++) {
   if (at->atm!=SAT->atm) continue;
   if (at->sym!=SAT->sym) continue;
   if (at->sft[0]!=SAT->sft[0]) continue;
   if (at->sft[1]!=SAT->sft[1]) continue;
   if (at->sft[2]!=SAT->sft[2]) continue;
   return 0;
  }
  return 1;
}

SV_Function void cmderr(char *a) 
{
  if (a!=NULL && (log_fp==NULL || fprintf(log_fp,"#%s\n",a)==EOF)) {
    fprintf(stderr,"Shelx View : Log file err.\n");
  }
  gputserr(a);
}

SV_Function void cmdinfo(char *a) 
{
  if (a!=NULL && (log_fp==NULL || fprintf(log_fp,"#%s\n",a)==EOF)) {
    fprintf(stderr,"Shelx View : Log file err.\n");
  }
  gputsinfo(a);
}

SV_Function void mat_mul(void)
{
  Double temp[3][3];
  int i,j,k;
  for(i=0;i<3;i++) {
    for(j=0;j<3;j++) {
      temp[i][j]=0;
      for(k=0;k<3;k++) {
        temp[i][j]+=matrix[i][k]*VMT[k][j];
      }
    }
  }
  for(i=0;i<3;i++) {
    for(j=0;j<3;j++) {
      VMT[i][j]=temp[i][j];
    }
  }
}

SV_Function void updlog(char *a)
{
  static FILE* ulog=NULL;
  if (ulog==NULL) {
    ulog=ffopen("sv.upd","w");
  }
  if (ulog!=NULL) fprintf(ulog,"%s",a);
}

SV_Function void setup_sym0(int sat_n,int at_n)
{
  S_AT *sat;
  ATOM_PARAM *at;
  sat=SAT+sat_n;
  at=orgAtom+at_n;
  sat->atm=at_n;
  sat->sym=0;
  sat->sft[0]=sat->sft[1]=sat->sft[2]=0;
  sat->frag=at->frag;
  sat->part=at->part;
  setup_symloc(sat,orgAtom);
  setup_atom(sat);
}

SV_Function void setup_symloc(S_AT *sat,ATOM_PARAM *base)
{
  int j,k;
  Float *x;
  x=sym_exp(sat,base);
  for(j=0;j<3;j++) {
    sat->vloc[j]=0;
    for(k=0;k<3;k++){
      sat->vloc[j]+=MAT[j][k]*x[k];
    }
  }
}

SV_Function int setup_atom(S_AT *sat) 
{
  char *nam;
  int j;
  ATOM_PARAM *oat;
  oat=orgAtom+sat->atm;
  nam=oat->name;
  if (oat->atom_type==0) {
    for(j=0;j<MAX_VDW;j++) {
      if (nam[0]!=vdw[j].name[0]) continue;
      if (isalpha(nam[1])) {
        if (nam[1]!=vdw[j].name[1]) continue;
      } else {
        if (vdw[j].name[1]!=' ') continue;
      };
      sat->radii=oat->radii=vdw[j].rad;
      sat->color=vdw[j].col;
      return 1;
    }
  } else {
    sat->radii=oat->radii=sfac[oat->atom_type-1].rad;
    sat->color=sfac[oat->atom_type-1].col;
  }
  return 0;
}

SV_Function void default_atom(ATOM_PARAM *atom,char *nam) 
{
  atom->uij[0]=0.06;
  atom->uij[1]=0;
  atom->occ=1;
  atom->peak_h=0;
  atom->afix=0;
  sprintf(atom->name,"%-4s",nam);
}

SV_Function S_AT *getatm(char *name)
{
  int i,j,flag,sm;
  int *s;
  Long adc;
  char *na;
  S_AT *sat;
  if (name[0]=='#') return SAT+atoi(name+1);
  adc=0;
  for(i=0;name[i]!=0;i++) {
    if (name[i]=='_') {
      adc=atol(name+i+1);
      name[i]=0;
    }
  }
  sat=DecodeAdc(adc<100?adc+55500:adc);
  sm=sat->sym;
  s=sat->sft;
  for(i=0;i<n_satom;i++) {
    na=OrgAtom(i)->name;
    flag=0;
    for(j=0;name[j]!=0;j++) {
      if (name[j]!=na[j]) flag=1;
    }
    if (na[j]!=0) flag=1;
    if (adc!=0) {
      if (adc>=100) {
        if (s[0]!=SAT[i].sft[0] || s[1]!=SAT[i].sft[1] 
            || s[2]!=SAT[i].sft[2]) flag=1;
      }
      if (sm!=SAT[i].sym) flag=1;
    }
    if (flag==0) {
      return SAT+i; 
    }
  }
  return NULL;
}

SV_Function Double *sym_exp(const S_AT *at1,const ATOM_PARAM *base)
{
  static Double loc[3];
  Float x[3];
  int i;
  for(i=0;i<3;i++) {
    x[i]=base[at1->atm].loc[i];
  }
  for(i=0;i<3;i++) {
    loc[i]=symm[at1->sym][i][0]+x[0]*symm[at1->sym][i][1]+
           x[1]*symm[at1->sym][i][2]+x[2]*symm[at1->sym][i][3]+
           at1->sft[i];
  };
  return loc;
}

SV_Function S_AT *get_length2(S_AT *at1,ATOM_PARAM *at2,Double std)
{
  static S_AT sat;
  int i,x30,x31,x32;
  Double x00,x01,x02  ,x10,x11,x12,  x20,x21,x22 ;
  Double x40,x41,x42  ,x50,x51,x52,  c_len,min_len,len;
  Double a0,a1,a2,a3,a4,a5;
  sat.atm=at2->num;
  std*=std;
  a0=MAT[0][0];
  a1=MAT[1][0]; a2=MAT[1][1];
  a3=MAT[2][0]; a4=MAT[2][1]; a5=MAT[2][2];
  min_len=1000;
  c_len=1000;
/* x0? = (vx1,vy1,vz1) */
  x00=at1->vloc[0];
  x01=at1->vloc[1];
  x02=at1->vloc[2];
/* x1? = (x2,y2,z2) */
  x10=at2->loc[0];
  x11=at2->loc[1];
  x12=at2->loc[2];
  for(i=0;i<symnum;i++) {
/* x2? = (x2,y2,z2) + symmetry */
    x20=symm[i][0][0] + symm[i][0][1]*x10
      + symm[i][0][2]*x11 + symm[i][0][3]*x12 ;
    x21=symm[i][1][0] + symm[i][1][1]*x10
      + symm[i][1][2]*x11 + symm[i][1][3]*x12 ;
    x22=symm[i][2][0] + symm[i][2][1]*x10
      + symm[i][2][2]*x11 + symm[i][2][3]*x12 ;
/* x3? = (trans x,trans y,trans z) */
    for (x30=SEARCH_MIN;x30<=SEARCH_MAX;x30++) {
    for (x31=SEARCH_MIN;x31<=SEARCH_MAX;x31++) {
    for (x32=SEARCH_MIN;x32<=SEARCH_MAX;x32++) {
/* x4? = (x2,y2,z2) + symmetry + translate */
      x40=x20+x30;
      x41=x21+x31;
      x42=x22+x32;
/* x5? = (vx2,vy2,vz2) */
      x50=a0*x40              -x00;
      x51=a1*x40+a2*x41       -x01;
      x52=a3*x40+a4*x41+a5*x42-x02;
      len=x50*x50+x51*x51+x52*x52;
      if (min_len>fabs(len-std)) {
        sat.sym=i;
        sat.sft[0]=x30;
        sat.sft[1]=x31;
        sat.sft[2]=x32;
	sat.atm=at2->num;
        min_len=fabs(len-std);
        c_len=len;
      }
    }}}
  }
  sat.radii=sqrt(c_len);
  return &sat;
}

/* to fix IRIX 6.2 buggy rint function */
SV_LocalFunction double Rint(double f)
{
  return floor(f+0.5);
}

SV_Function Float get_length3(ATOM_PARAM *at1,ATOM_PARAM *at2)
{
  int i,x30,x31,x32;
  Double x00,x01,x02  ,x10,x11,x12,  x20,x21,x22 ;
  Double x40,x41,x42  ,x50,x51,x52,  c_len,min_len,m_len,len;
  Double a0,a1,a2,a3,a4,a5;
  c_len=at1->radii+at2->radii;
  c_len*=c_len;
  a0=MAT[0][0];
  a1=MAT[1][0]; a2=MAT[1][1];
  a3=MAT[2][0]; a4=MAT[2][1]; a5=MAT[2][2];
  min_len=1000;
/* x0? = (x1,y1,z1) */
  x00=at1->loc[0]-Rint(at1->loc[0]);
  x01=at1->loc[1]-Rint(at1->loc[1]);
  x02=at1->loc[2]-Rint(at1->loc[2]);
/* x1? = (x2,y2,z2) */
  x10=at2->loc[0]-floor(at2->loc[0]);
  x11=at2->loc[1]-floor(at2->loc[1]);
  x12=at2->loc[2]-floor(at2->loc[2]);
  for(i=0;i<symnum;i++) {
/* x2? = (x2,y2,z2) + symmetry op */
    x20=symm[i][0][0] + symm[i][0][1]*x10
      + symm[i][0][2]*x11 + symm[i][0][3]*x12 ;
    x21=symm[i][1][0] + symm[i][1][1]*x10
      + symm[i][1][2]*x11 + symm[i][1][3]*x12 ;
    x22=symm[i][2][0] + symm[i][2][1]*x10
      + symm[i][2][2]*x11 + symm[i][2][3]*x12 ;
/* x3? = (trans x,trans y,trans z) */
    for (x30=-1;x30<=1;x30++) {
    for (x31=-1;x31<=1;x31++) {
    for (x32=-1;x32<=1;x32++) {
/* x4? = (x2,y2,z2) + symmetry + translate - (x1,y1,z1)*/
      x40=x20+x30-x00;
      x41=x21+x31-x01;
      x42=x22+x32-x02;
/* x5? = (vx2,vy2,vz2) */
      x50=a0*x40;
      x51=a1*x40+a2*x41;
      x52=a3*x40+a4*x41+a5*x42;
      len=x50*x50+x51*x51+x52*x52;
      if (len>16) continue;
      if (min_len>sqrt(fabs(len-c_len))) {
        min_len=sqrt(fabs(len-c_len));
	m_len=len;
      }
    }}}
  }
  return m_len;
}

SV_Function Double get_length(const S_AT *at1,const S_AT *at2)
{
  Double dx,dy,dz;
  dx=at1->vloc[0]-at2->vloc[0];
  dy=at1->vloc[1]-at2->vloc[1];
  dz=at1->vloc[2]-at2->vloc[2];
  return sqrt(dx*dx+dy*dy+dz*dz);
}

SV_Function Double get_angle(const S_AT *at1,const S_AT *at2,const S_AT *at3)
{
  Double dx12,dy12,dz12,dx13,dy13,dz13,dx23,dy23,dz23,d12,d23,d13;
  Double t1,t2;
  dx12=at1->vloc[0]-at2->vloc[0];
  dy12=at1->vloc[1]-at2->vloc[1];
  dz12=at1->vloc[2]-at2->vloc[2];
  dx13=at1->vloc[0]-at3->vloc[0];
  dy13=at1->vloc[1]-at3->vloc[1];
  dz13=at1->vloc[2]-at3->vloc[2];
  dx23=at3->vloc[0]-at2->vloc[0];
  dy23=at3->vloc[1]-at2->vloc[1];
  dz23=at3->vloc[2]-at2->vloc[2];
  d12=dx12*dx12+dy12*dy12+dz12*dz12;
  d23=dx23*dx23+dy23*dy23+dz23*dz23;
  d13=dx13*dx13+dy13*dy13+dz13*dz13;
  if (d12*d23==0) return 0;
  t1=d12+d23-d13;
  t2=sqrt(d12*d23*4);
  if ((t1/t2>1.0000 && t1/t2<1.00001) || (t1/t2<-1.0000 && t1/t2>-1.00001)) {
    t1=t2=1.0;
  }
  return acos(t1/t2)/PI*180;
}

SV_Function Double get_torsion(const S_AT *at1,const S_AT *at2,const S_AT *at3,const S_AT *at4)
{
  Double dx12,dy12,dz12,dx23,dy23,dz23,dx34,dy34,dz34,t1,t2,t3;
  dx12=at2->vloc[0]-at1->vloc[0];
  dy12=at2->vloc[1]-at1->vloc[1];
  dz12=at2->vloc[2]-at1->vloc[2];
  dx23=at3->vloc[0]-at2->vloc[0];
  dy23=at3->vloc[1]-at2->vloc[1];
  dz23=at3->vloc[2]-at2->vloc[2];
  dx34=at4->vloc[0]-at3->vloc[0];
  dy34=at4->vloc[1]-at3->vloc[1];
  dz34=at4->vloc[2]-at3->vloc[2];
  t1=(
     (dy12*dz23-dy23*dz12)*(dy23*dz34-dy34*dz23) +
     (dz12*dx23-dz23*dx12)*(dz23*dx34-dz34*dx23) +
     (dx12*dy23-dx23*dy12)*(dx23*dy34-dx34*dy23)
  );
  t2=(
     sqrt(dx12*dx12+dy12*dy12+dz12*dz12)*
     ( dx23*dx23+dy23*dy23+dz23*dz23 )*
     sqrt(dx34*dx34+dy34*dy34+dz34*dz34) *
     sin(get_angle(at1,at2,at3)*PI/180) *
     sin(get_angle(at2,at3,at4)*PI/180)
  );
  if (t2==0) return 360;
  t3=(dx12*((dy23*dz34)-(dy34*dz23))+
      dy12*((dz23*dx34)-(dz34*dx23))+
      dz12*((dx23*dy34)-(dx34*dy23)))/t2;
  if ((t1/t2>1.0000 && t1/t2<1.00001) || (t1/t2<-1.0000 && t1/t2>-1.00001)) {
    t1=t2=1.0;
  }
  return 180.0/PI*acos(t1/t2)*(t3>0?1:-1);
}

SV_Function void putstr(void)
{
  int i;
  gcolor(7);
  if (disp_str[0][0]==0)  return;
  gputs2("");
  for(i=0;disp_str[i][0]!=0;i++)
    gputs2(disp_str[i]);
  gputs2("#");
  disp_str[0][0]=0;
}

SV_Function int IsNonPosi(double *U)
{
    if (U[0]<0 || U[1]<0 || U[2]<0 ||
       U[0]*U[1]-U[5]*U[5]<0 || U[1]*U[2]-U[3]*U[3]<0 || U[2]*U[0]-U[4]*U[4]<0 ||
       U[0]*U[1]*U[2]+2*U[3]*U[4]*U[5]-U[2]*U[5]*U[5]-U[0]*U[3]*U[3]
              -U[4]*U[1]*U[4]<0 ) return 1;

return 0;
}

SV_Function ATOM_PARAM *NewAtom(void)
{
  if (n_atom+1>=MAX_ATOM) {
    orgAtom=(ATOM_PARAM *)Realloc(orgAtom,&MAX_ATOM,ATOM_BLOCK_SIZE,
                                  sizeof(ATOM_PARAM));
  }
  orgAtom[n_atom].num=n_atom;
  orgAtom[n_atom+1].num=-1;
  return orgAtom+(n_atom++);
}

SV_Function S_AT *NewSat(void)
{
  if (n_satom+1>=MAX_SYMATOM) {
    SAT=(S_AT *)Realloc(SAT,&MAX_SYMATOM,ATOM_BLOCK_SIZE,sizeof(S_AT));
  }
  SAT[n_satom].num=n_satom;
  SAT[n_satom+1].num=-1;
  return SAT+(n_satom++);
}

SV_Function BOND_PARAM *NewBond(void)
{
  if (n_bond+1>=MAX_BOND) {
    BOND=(BOND_PARAM *)
      Realloc(BOND,&MAX_BOND,ATOM_BLOCK_SIZE,sizeof(BOND_PARAM));
    if (BOND==NULL) {
      PutMessage1("Out of Memory!");
      Exit(1);
    }
  }
  BOND[n_bond].num=n_bond;
  BOND[n_bond+1].num=-1;
  return BOND+(n_bond++);
}
