/* Cmd SubModule  "view.c"           -*- C -*- */

/*
#S#
#
# This Program is part of SV (Shelx-output Viewer)
# Copyright (C) 1991-1998 Takashi Nemoto
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details. (File "COPYING")
#
#    Note!: General Public Licence is copyrighted by the Free Software 
#           Foundation,  but this code is copyrighted by me.
#
#    Send bugs and comments to tnemoto@eels.kuicr.kyoto-u.ac.jp
#
#E#
*/

#include "config.h"
#define __view_c__
#define __cmd_c_proc__
#define __subs_c_proc__
#define __stack_c_proc__
#define __shelplot_c_proc__
#define __bond_c_proc__
#ifdef Xt
#define __xtcalls_c_proc__
#endif

#include <math.h>
#include <ctype.h>
#include "def.h"

SV_Function void view(void)
{
  int i,j,k;
  S_AT *at;
  char buf[30];
  char buf2[30];
  switch (cmd_c) {
  case 1:
    if (chkinst(cmd_param[0],"SL","SELECT")) {
      sprintf(disp_str[0],"Selected Atom list");
      sprintf(disp_str[1]," ");
      j=2;
      i=0;
      k=0;
      buf[0]=0;
      for(i=atom_stack_top;i>=0;i--) {
        sprintf(buf2,"%s %-4s",buf,orgAtom[SAT[atom_stack[i]].atm].name);
        sprintf(buf,"%s",buf2);
        k++;
        if (k==4) {
          sprintf(disp_str[j++],"%s",buf);
          k=0;
          buf[0]=0;
        }
      }
      sprintf(disp_str[j++],"%s",buf);
      disp_str[j++][0]=0;
      refresh();
      return;
    }
    at=cmd_atom[0];
    if (at!=NULL) {
/*      for(i=0;i<2;i++) shift[i]+=at->axy[i]/vscale;*/
      for(i=0;i<3;i++) offst[i]=at->vloc[i];
      global_flag&=~GLOBALFLAG_GC;
      shift0();
      return ;
    }
    if (chkinst(cmd_param[0],"gc","gcent")!=0) {
      global_flag|=GLOBALFLAG_GC;
    }
    if (cmd_param[0][1]!=0) return ;
    switch (cmd_param[0][0]) {
    case '0':
      VMT[0][0]=1;VMT[0][1]=0;VMT[0][2]=0;
      VMT[1][0]=0;VMT[1][1]=1;VMT[1][2]=0;
      VMT[2][0]=0;VMT[2][1]=0;VMT[2][2]=1;
      refresh();
      return ;
    case 'A':
    case 'X':
      cmd_c=2;
      cmd_atom[1]=SYMCELL+0;
      cmd_atom[0]=SYMCELL+1;
      view();
      return;
    case 'B':
    case 'Y':
      cmd_c=2;
      cmd_atom[1]=SYMCELL+0;
      cmd_atom[0]=SYMCELL+2;
      view();
      return;
    case 'C':
    case 'Z':
      cmd_c=2;
      cmd_atom[1]=SYMCELL+0;
      cmd_atom[0]=SYMCELL+4;
      view();
      return;
    default:
      cmderr("## vw : unknown param\n");
      return ;
    }
  case 2:
    view_line();
    break;
  case 3:
    if (cmd_num[0]!=0 || cmd_num[1]!=0 || cmd_num[2]!=0) {
      view_axis();
    } else {
      cmd_atom[4]=cmd_atom[0];
      cmd_atom[0]=cmd_atom[2];
      view_line();
      cmd_atom[0]=cmd_atom[4];
      view_line();
    }
    break;
    /*
  case 4:
    if (chkinst(cmd_param[0],"PL","PLANE") && 
	(cmd_num[1]!=0  || cmd_num[2]!=0 || cmd_num[3]!=0)) {
      view_plane1();
      break;
    }
    */
  default:
    cmderr("## vw : paramater error\n");
    break;
  }
  refresh();
}


SV_Function void view_line(void)
{
  int i,j,k;
  Double xyz[2][3],th1,th2;
  if (cmd_atom[0]==NULL || cmd_atom[1]==NULL) {
    char buf[cBufSize];
    sprintf(buf,"Atom not found: 'vw %s %s' ignored",
	    cmd_param[0],cmd_param[1]);
    PutWarning1(buf);
    return;
  }
  for(i=0;i<2;i++)
    {
      for(j=0;j<3;j++)
        {
          xyz[i][j]=0;
          for(k=0;k<3;k++)
            {
              xyz[i][j]+=VMT[j][k]*cmd_atom[i]->vloc[k];
            }
        }
    }
  for(i=0;i<3;i++)
    {
      xyz[1][i]=xyz[0][i]-xyz[1][i];
    }
  th1=atan2(xyz[1][1],xyz[1][0]);
  th2=atan2(sqrt(xyz[1][1]*xyz[1][1]+xyz[1][0]*xyz[1][0]),xyz[1][2]);
  rotz(-th1);
  roty(th2);
}

SV_LocalFunction void view_axis(void)
{
  int i,j;
  Double xyz[3],xyz2[3],th1,th2;
  for(i=0;i<3;i++) {
    xyz[i]=0;
    for(j=0;j<3;j++){
      xyz[i]+=MAT[i][j]*cmd_num[j];
    }
  }
  for(i=0;i<3;i++) {
    xyz2[i]=0;
    for(j=0;j<3;j++) {
      xyz2[i]+=VMT[i][j]*xyz[j];
    }
  }
  th1=atan2(xyz2[1],xyz2[0]);
  th2=atan2(sqrt(xyz2[1]*xyz2[1]+xyz2[0]*xyz2[0]),xyz2[2]);
  rotz(-th1);
  roty(th2);
}

SV_Function void view_plane(int n_loc,Double (*xyz)[3]) {
  Double ave[3];
  Double rr[3],rrP[3];
  Double sft=0,delsft=0,risk=0;
  Double pla,plb,plc,pld,cplb,cplc;
  int ncycle;
  int i,j;
  int ncon=0;
  if (n_loc<3) return;
  ave[0]=ave[1]=ave[2]=0;
  for(i=0;i<n_loc;i++){
    for(j=0;j<3;j++){
      ave[j]+=xyz[j][i];
    }
  }
  for(i=0;i<3;i++) ave[i]/=n_loc;
  for(i=0;i<n_loc;i++){
    for(j=0;j<3;j++){
      xyz[j][i]-=ave[i];
    }
  }
  for(i=0;i<n_loc;i++){
    for(j=0;j<3;j++){
      rr[j]+=xyz[j][i]*xyz[j][i];
      rrP[j]+=xyz[(j+1)%3][i]*xyz[(j+2)%3][i];
    }
  }
  if (n_loc!=3) {
    Double coef[4];
    coef[0]=-1;
    coef[1]=rr[0]+rr[1]+rr[2];
    coef[2]=rrP[0]*rrP[0]+rrP[1]*rrP[1]+rrP[2]*rrP[2]
      -(rr[0]*rr[1]+rr[1]*rr[2]+rr[2]*rr[0]);
    coef[3]=rr[0]*rr[1]*rr[2]+2*rrP[0]*rrP[1]*rrP[2]
      -rr[0]*rrP[0]*rrP[0]-rr[1]*rrP[1]*rrP[1]-rr[2]*rrP[2]*rrP[2];
    do {
      Double diff=0;
      Double func;
      func=coef[0];
      for(j=1;j<4;j++){
	diff=diff*sft+func;
	func=func*sft+coef[j];
      }
      if (fabs(diff)<1.0E-20) break;
      delsft=-func/diff;
      sft=sft+delsft;
      ncycle=ncycle+1;
    } while (fabs(delsft)>0.1E-7);
  }
  rr[0]-=sft;
  rr[1]-=sft;
  risk=rr[0]*rr[1]-rrP[2]*rrP[2];
  if (risk>1.0E-20) {
    pla=(rrP[2]*rrP[0]-rr[1]*rrP[1])/risk;
    plb=(rrP[2]*rrP[1]-rr[0]*rrP[0])/risk;
    plc=1.0/sqrt(1.0+pla*pla+plb*plb);
    pla=pla*plc;
    plb=plb*plc;
  } else {
    plc=0.0;
    if (rr[0]>1.0E-20) {
      pla=-rrP[2]/rr[0];
      plb=1.0/sqrt(1.0+pla*pla);
      pla=pla*plb;
    } else {
      pla=1.0;
      plb=0.0;
    }
  }
  if (ncon) {
    pld=0.0;
  } else {
    pld=-(ave[0]*pla+ave[1]*plb+ave[2]*plc);
    if (pld<0) {
      pla=-pla;
      plb=-plb;
      plc=-plc;
      pld=-pld;
    }
  }
  cplb=pla*cell_cos[2]+plb*cell_sin[2];
  cplc=pla*cell_cos[1]-plb*cell_cos_s[0]*cell_sin[1]+plc/(cell[2]*rcell[2]);
  sprintf(disp_str[0],"Least Square Plane");
  sprintf(disp_str[1],"%6.2fx+%6.2fy+%6.2fz=%6.2f",pla,cplb,cplc,pld);
  disp_str[2][0]=0;
  refresh();
  /*
  for(j=0;j<NSFTA;j++){
    sftatm=pla*plx(j)+plb*ply(j)+plc*plz(j);
    mm=natm(j);
    WRITE (3,44) ATOM(MM),NUN(J),SFTATM;
  }
  */
}

SV_Function void depth(void)
{
  if (cmd_param[0][0]=='A' && cmd_c==2) ZVIEW=cmd_num[1]; 
  else if (cmd_c==1) ZVIEW+=cmd_num[0];
  if (ZVIEW<0) ZVIEW=0;
#ifdef Xt
  DispDepth();
#endif
  refresh();
}

SV_Function void rot_step_set(void)
{
  if (cmd_param[0][0]=='A' && cmd_c==2) rot_step=cmd_num[1]; 
  else if (cmd_c==1) rot_step+=cmd_num[0];
  if (rot_step>360 || rot_step <0) rot_step=0;
}

SV_Function void rot(void)
{
  if (cmd_c!=3) return;
  rotx(cmd_num[0]);
  roty(cmd_num[1]);
  rotz(cmd_num[2]);
  refresh();
}

SV_Function void rot_x(void)
{
  rotx(cmd_num[0]*PI/180);
  refresh();
}

SV_Function void rotx(Double theta)
{
  Double ct,st;
  st=sin(theta);
  ct=cos(theta);
  matrix[0][0]=1; matrix[0][1]=0; matrix[0][2]=0;
  matrix[1][0]=0; matrix[1][1]=ct; matrix[1][2]=st;
  matrix[2][0]=0; matrix[2][1]=-st; matrix[2][2]=ct;
  mat_mul();
}

SV_Function void rot_y(void)
{
  roty(cmd_num[0]*PI/180);
  refresh();
}

SV_Function void roty(Double theta)
{
  Double ct,st;
  st=sin(theta);
  ct=cos(theta);
  matrix[0][0]=ct; matrix[0][1]=0; matrix[0][2]=-st;
  matrix[1][0]=0; matrix[1][1]=1; matrix[1][2]=0;
  matrix[2][0]=st; matrix[2][1]=0; matrix[2][2]=ct;
  mat_mul();
}

SV_Function void rot_z(void)
{
  rotz(cmd_num[0]*PI/180);
  refresh();
}

SV_Function void rotz(Double theta)
{
  Double ct,st;
  st=sin(theta);
  ct=cos(theta);
  matrix[0][0]=ct; matrix[0][1]=-st; matrix[0][2]=0;
  matrix[1][0]=st; matrix[1][1]=ct; matrix[1][2]=0;
  matrix[2][0]=0; matrix[2][1]=0; matrix[2][2]=1;
  mat_mul();
}

SV_Function void shift_x(void)
{
  shift[0]+=cmd_num[0];
  refresh();
}

SV_Function void shift_y(void)
{
  shift[1]+=cmd_num[0];
  refresh();
}

SV_Function void shift0(void)
{
  shift[0]=shift[1]=0;
  refresh();
}

SV_Function void omit_view(void)
{
  int i,j,k,fl;
  for(k=0;k<cmd_c;k++){
    if (chkinst(cmd_param[k],"BL","BONDLEN")) {
      global_flag&=~GLOBALFLAG_BONDLEN;
    } else if (chkinst(cmd_param[k],"AF","AFIX")) {
      global_flag&=~GLOBALFLAG_DISPAFIX;
    } else if (chkinst(cmd_param[k],"DF","DFIX")) {
      global_flag&=~GLOBALFLAG_DISPDFIX;
    } else if (chkinst(cmd_param[k],"LB","LABEL")) {
      if (k+1<cmd_c && chkinst(cmd_param[k+1],"SL","SELECT")){
	for(i=0;i<=atom_stack_top;i++){
	  SAT[atom_stack[i]].view_flag |=FLAG_NOLABEL;
	}
	stack_clr();
	k++;
      } else {
	global_flag&=~GLOBALFLAG_LABEL;
      }
    } else if (chkinst(cmd_param[k],"LL","LONGLABEL")){
      global_flag&=~GLOBALFLAG_LONGLABEL;
    } else if (chkinst(cmd_param[k],"CELL","CELL")) {
      global_flag&=~GLOBALFLAG_CELL;
    } else if (chkinst(cmd_param[k],"SL","SELECT")) {
      for(i=0;i<=atom_stack_top;i++){
        SAT[atom_stack[i]].view_flag|=FLAG_OMIT;
      }
      stack_clr();
    } else if (chkinst(cmd_param[k],"FR","FRAGMENT") && k<cmd_c-1) {
      k++;
      for(i=0;i<n_satom;i++){
	if (SAT[i].frag==cmd_int[k]) {
	  SAT[i].view_flag|=FLAG_OMIT;	  
	}
      }
    } else if (cmd_param[k][1]==0 || 
             (cmd_param[k][2]==0 && !isdigit(cmd_param[k][1]))) {
      for(i=0;i<n_satom;i++) {
        fl=0;
        for(j=0;cmd_param[k][j]!=0;j++) {
          if (OrgAtom(i)->name[j]!=cmd_param[k][j]) fl=1;
        }
        if (fl==0)SAT[i].view_flag|=FLAG_OMIT;
      }
    } else {
      if (cmd_atom[k]!=NULL) cmd_atom[k]->view_flag|=FLAG_OMIT;
    }
  }
  refresh();
}

SV_Function void purge(void)
{
  int i;
  for(i=0;i<n_satom;i++){
    if ((SAT[i].view_flag & FLAG_OMIT)!=0) {
      SAT[i]=SAT[--n_satom];
    }
  }
  InsertCommandQueue("makebond auto");
}

SV_Function void include_view(void)
{
  int i,j,k,fl;
  for(k=0;k<cmd_c;k++){
    if (chkinst(cmd_param[k],"BL","BONDLEN")) {
      global_flag|=GLOBALFLAG_BONDLEN;
    } else if (chkinst(cmd_param[k],"AF","AFIX")) {
      global_flag|=GLOBALFLAG_DISPAFIX;
    } else if (chkinst(cmd_param[k],"DF","DFIX")) {
      global_flag|=GLOBALFLAG_DISPDFIX;
    } else if (chkinst(cmd_param[k],"LB","LABEL")) {
      if (k+1<cmd_c && chkinst(cmd_param[k+1],"SL","SELECT")){
	for(i=0;i<=atom_stack_top;i++){
	  SAT[atom_stack[i]].view_flag &=~ FLAG_NOLABEL;
	}
	stack_clr();
	k++;
      } else {
	global_flag|=GLOBALFLAG_LABEL;
      }
    } else if (chkinst(cmd_param[k],"LL","LONGLABEL")){
      global_flag|=GLOBALFLAG_LABEL | GLOBALFLAG_LONGLABEL;
    } else if (chkinst(cmd_param[k],"CELL","CELL")) {
      global_flag|=GLOBALFLAG_CELL;
    } else if (chkinst(cmd_param[k],"SL","SELECT")) {
      for(i=0;i<=atom_stack_top;i++){
        SAT[atom_stack[i]].view_flag&=~FLAG_OMIT;
      }
      stack_clr();
    } else if (chkinst(cmd_param[k],"FR","FRAGMENT") && k<cmd_c-1) {
      k++;
      for(i=0;i<n_satom;i++){
	if (SAT[i].frag==cmd_int[k]) {
	  SAT[i].view_flag&=~FLAG_OMIT;	  
	}
      }
    } else if (cmd_param[k][1]==0 || 
             (cmd_param[k][2]==0 && !isdigit(cmd_param[k][1]))) {
      for(i=0;i<n_satom;i++) {
        fl=0;
        for(j=0;cmd_param[k][j]!=0;j++) {
          if (OrgAtom(i)->name[j]!=cmd_param[k][j]) fl=1;
        }
        if (fl==0)SAT[i].view_flag&=~FLAG_OMIT;
      }
    } else {
      if (cmd_atom[k]!=NULL) cmd_atom[k]->view_flag&=~FLAG_OMIT;
    }
  }
  refresh();
}

SV_Function void peak(void) 
{
  int min,max,i,q0=-1,j=0;
  min=1;
  if (cmd_c==0) max=999;
  else if (cmd_c==1) max=cmd_int[0];
  else { min=cmd_int[0]; max=cmd_int[1]; };
  for(i=0;i<n_atom;i++) { 
    if (orgAtom[i].name[0]=='Q') {
      q0=i; break;
    }
  }
  if (q0<0) return;
  sprintf(disp_str[j++],"Peak List");
  sprintf(disp_str[j++]," ");
  for(i=q0;i<n_satom;i++) {
    SAT[i].view_flag|=FLAG_OMIT;
  };
  for(i=q0+min-1;i<q0+max;i++) {
    if (i<n_satom && i>=0) {
      SAT[i].view_flag&=~FLAG_OMIT;
      if (j<MAX_LINE-1) {
        sprintf(disp_str[j++],"%-4s %5.2f %7.4f%7.4f%7.4f",orgAtom[i].name,
                orgAtom[i].peak_h,orgAtom[i].loc[0],orgAtom[i].loc[1],
                orgAtom[i].loc[2]);
      }
    }
  }
  disp_str[j][0]=0;
  refresh();
}

SV_Function void scale_view(void)
{
  if (cmd_num[0]>0) vscale*=cmd_num[0];
  refresh();
}

