/* Graphics Library Module   "shelplot.c"    -*- C -*-  */

/*
#S#
#
# This Program is part of SV (Shelx-output Viewer)
# Copyright (C) 1991-1993 Takashi Nemoto
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details. (File "COPYING")
#
#    Note!: General Public Licence is copyrighted by the Free Software 
#           Foundation,  but this code is copyrighted by me.
#
#    Send bugs and comments to tnemoto@eels.kuicr.kyoto-u.ac.jp
#
#E#
*/
/*
 * This file contains

 void ginit(void)
 void gcolor(int col)
 void gopen(void)
 void gcls(void)
 void gline(Double x,Double y,Double x2,Double y2)
 void gputs(Double x,Double y,char *txt)
 void gputserr(char *txt)
 void gputsinfo(char *txt)
 void gputs2(char *txt)
 void gputscmd(void);
 void gclose(void);
 void ginit2(void);
 void gflush(void);
 void dispgraph(char *txt,int num);
 void gdisp_atom_stack(void);

*/

#define __shelplot_c__
#define __cmd_c_proc__
#define __stack_c_proc__
#define __winmain_c_proc__
#define __wincalls_c_proc__
#include <string.h>
#include "../common/def.h"
#include <windows.h>
#include <commdlg.h>
#include <commctrl.h>
#include "winmain.h"
extern  HANDLE MyInstance;
extern   HWND        hwndMainWin,hwndViewWin,hwndCmdWin,hwndListWin,hwndStackWin;                                                  
extern   HFONT  hfViewWin;

extern TCHAR fontname[LF_FACESIZE];
extern LONG fontsize;

#define  ListWinBufLength 10000
char textbuf[2][ListWinBufLength];

SV_Variable int MAX_LINE
                        =ListWinBufLength/50;
SV_Variable int clsFlag
                       =1;
HDC hdcViewWin;
HPEN hpenViewWin[33];

SV_Variable Float winXScale,winYScale,winXCenter,winYCenter;

#define FNX(x) ((x)*winXScale+winXCenter)
#define FNY(y) (-(y)*winYScale+winYCenter)

const int svGray[]={0,16,32,48,64,80,96,112,128,144,160,176,192,208,224,240};
COLORREF colorRef[32];
RECT rectViewWin;
char *errorText;

SV_Function void WinRefresh(void)
{
    SelectObject(hdcViewWin,hfViewWin);
    refresh();
} 

SV_Function void WinPaint(void)
{
    HDC orgHDC;
    PAINTSTRUCT ps;
    orgHDC=hdcViewWin;
    hdcViewWin=BeginPaint(hwndViewWin,&ps);
    WinRefresh();
    EndPaint(hwndViewWin,&ps);
    hdcViewWin=orgHDC;
} 

int ColorPlot=1;
SV_Function void ginit(void)
{
  int i;
  colorRef[0]=RGB(0,0,0);
  colorRef[1]=RGB(0,0,255);
  colorRef[2]=RGB(255,0,0);
  colorRef[3]=RGB(255,0,255);
  colorRef[4]=RGB(0,255,0);
  colorRef[5]=RGB(0,255,255);
  colorRef[6]=RGB(255,255,0);
  colorRef[7]=RGB(255,255,255);
  for(i=0;i<8;i++) colorRef[i+8]=colorRef[i];
  for(i=0;i<16;i++) colorRef[i+16]=RGB(svGray[i],svGray[i],svGray[i]);
  for(i=0;i<32;i++)hpenViewWin[i]=CreatePen(PS_SOLID,1,colorRef[i]);
  hpenViewWin[32]=CreatePen(PS_DOT,1,RGB(0,0,0));
  gcolor(0);
  ColorPlot=1;
}

SV_Function void LPTinit(void)
{
  int i;
  if (global_flag & GLOBALFLAG_COLORPLOT) {
    ColorPlot=1;
  } else {
    ColorPlot=0;
  }
/*
  int i;
    colorRef[0]=RGB(0,0,0);
    colorRef[1]=RGB(0,0,127);
    colorRef[2]=RGB(127,0,0);
    colorRef[3]=RGB(127,0,127);
    colorRef[4]=RGB(0,127,0);
    colorRef[5]=RGB(0,127,127);
    colorRef[6]=RGB(127,127,0);
    colorRef[7]=RGB(255,255,255);
    for(i=0;i<8;i++) colorRef[i+8]=colorRef[i];
    for(i=0;i<16;i++) colorRef[i+16]=RGB(svGray[i]/2,
					 svGray[i]/2,svGray[i]/2);
    for(i=0;i<32;i++)hpenViewWin[i]=CreatePen(PS_SOLID,1,colorRef[i]);
*/

    gcolor(0);
}

SV_Function void gcolor(int col)
{
  static int oldColor=-1;
  if (ColorPlot==0) col=0;
  if (col==oldColor) return;
  if (col<0 || col>32) return;
  if (col==32) {
    SelectObject(hdcViewWin,hpenViewWin[col]);
    return;
  }
  if (col<8 || col>=16) {
    SetTextColor(hdcViewWin,colorRef[col]);
    SetBkColor(hdcViewWin,colorRef[7]);
    SetBkMode(hdcViewWin,TRANSPARENT);
  } else {
    SetTextColor(hdcViewWin,colorRef[7]);
    SetBkColor(hdcViewWin,colorRef[col]);
    SetBkMode(hdcViewWin,OPAQUE);
  }
  SelectObject(hdcViewWin,hpenViewWin[col]);
}

SV_Function void gopen(void)
{
  int x0,x1,y1;
  hdcViewWin=GetDC(hwndViewWin);
  SelectObject(hdcViewWin,hfViewWin);
  GetClientRect(hwndViewWin,&rectViewWin);
  x1=rectViewWin.right;
  y1=rectViewWin.bottom;
  winXCenter=x1>>1;
  winYCenter=y1>>1;
  x0=min(x1,y1);
  winXScale=winYScale=0.1*x0;
}

SV_Function void gcls(void)
{
  HBRUSH hBrush;
  hBrush = CreateSolidBrush(RGB(255,255,255));
  FillRect(hdcViewWin, &rectViewWin,hBrush);
  DeleteObject(hBrush);
  clsFlag=1;
}

SV_Function void gline(Double x,Double y,Double x2,Double y2)
{
    MoveToEx(hdcViewWin,FNX(x),FNY(y),NULL);
    LineTo(hdcViewWin,FNX(x2),FNY(y2));
}

SV_Function void gputs(Double x,Double y,char *txt)
{
    TextOut(hdcViewWin,FNX(x)-3,FNY(y)-5,txt,strlen(txt));
}

SV_Function void gputserr(char *txt)
{
        if (txt==NULL) return;
        MessageBeep(MB_ICONEXCLAMATION);
        MessageBox(NULL,txt,"Error",MB_ICONEXCLAMATION | MB_OK |MB_TASKMODAL);
        SetFocus(hwndCmdWin);
}

SV_Function void gputsinfo(char *txt)
{
        if (txt==NULL) return;
        MessageBeep(MB_ICONINFORMATION);
        MessageBox(NULL,txt,"Message:",MB_ICONINFORMATION | MB_OK |MB_TASKMODAL);
        SetFocus(hwndCmdWin);
}

void gputsw(HANDLE hwnd,char *txt,int bufnum) {
        if (txt[0]==0) {
                SendMessage(hwnd, EM_SETSEL, 0, -1);
                SendMessage(hwnd,WM_CLEAR, 0, 0L);
		textbuf[bufnum][0]=0;
        } else if (txt[0]=='#') {
                SetWindowText(hwnd,textbuf[bufnum]);
                return;
        } else {
		int len;
		len=strlen(txt);
                if (strlen(textbuf[bufnum])>ListWinBufLength-len-10) {
                    return;
                }

                strcat(textbuf[bufnum],txt);
                strcat(textbuf[bufnum],"\r\n");

                if (strlen(textbuf[bufnum])>ListWinBufLength-100) {
		  strcat(textbuf[bufnum],"\r\n=== over flow ===\r\n");
		}
		//                SetWindowText(hwnd,textbuf[bufnum]);
        }
}

SV_Function void gputs2(char *txt)
{
    gputsw(hwndListWin,txt,0);
}

void gputs3(char *txt)
{
    gputsw(hwndStackWin,txt,1);
}

SV_Function void gputscmd(void)
{}

SV_Function void gclose(void)
{
  ReleaseDC(hwndViewWin,hdcViewWin);
}

SV_Function void gflush(void)
{
    UpdateWindow(hwndViewWin);
}

BOOL CALLBACK ProgressFunc(HWND hdwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
  switch(msg){
  case WM_INITDIALOG:
    return 1;
  }
  return 0;
}

HWND hwndProgress=NULL;
SV_Function void dispgraph(char *txt,int num)
{
  if (num==-1 && hwndProgress==NULL){
    hwndProgress=CreateDialog(MyInstance,"PROGRESS",hwndViewWin,ProgressFunc);
  } else if (num==-2) {
    DestroyWindow(hwndProgress);
    hwndProgress=NULL;
    SetFocus(hwndCmdWin);
  }
  if (hwndProgress==NULL) return;
  SendMessage(hwndProgress,WM_SETTEXT,0,(LPARAM)(LPCTSTR)txt);
  SendDlgItemMessage(hwndProgress,PROGRESS_BAR,PBM_SETPOS,(WPARAM)num,0);
/*
  double rx,ry;
  if (txt==NULL) return;
  
  rx=0.05*(100-num)*winXScale;
  ry=0.05*(100-num)*winXScale;
  gcolor(5);
  Ellipse(hdcViewWin,winXCenter-rx,winYCenter-ry,winXCenter+rx,winYCenter+ry);
  gcolor(0);
  gputs(-0.1,0,txt);        
*/
}

SV_Function void gdisp_atom_stack(void)
{
    char txt1[50],txt2[50];

    int i;
    gputs3("");
    gputs3("Atom  Bond  Angle  Torsion");

    for(i=0;i<=(atom_stack_top>9?9:atom_stack_top);i++){
      sprintf(txt1,"%-4s",OrgAtom(atom_stack_s[i]->atm)->name);
      if (i>0) {
        sprintf(txt2," %-6.2f",atom_st_bnd[i-1]);
        strcat(txt1,txt2);
        if (i>1) {
          sprintf(txt2," %-6.2f",atom_st_ang[i-2]);
          strcat(txt1,txt2);
          if (i>2){
            sprintf(txt2,"  %-6.2f",atom_st_tor[i-3]);
            strcat(txt1,txt2);
          }
        }
      }
      gputs3(txt1);
    }

    gputs3("#");
}


HDC lptHDC=NULL;
HFONT hfLPT;
DOCINFO di;
PRINTDLG pd;

HDC emfHDC=NULL;
HFONT hfEMF;

SV_Function Bool LPTopen(void)
{
  static int RecursiveCall=0;
  int x0,x1,y1,i;

  /* To avoid recursive call in refresh action during PrintDlg is opened */
  if (RecursiveCall!=0) return;  

  memset((void *)&pd,0,sizeof(PRINTDLG));
  pd.lStructSize=sizeof(PRINTDLG);
  pd.hwndOwner=hwndMainWin;
  pd.Flags=PD_RETURNDC;
  pd.hInstance=NULL;
  RecursiveCall=1;

  if (PrintDlg(&pd)!=TRUE) {
    /*
      gputserr("PrintDlg Error.");
    */
    RecursiveCall=0;
    return ERR;
  }
  RecursiveCall=0;

  if (pd.hDevMode) GlobalFree(pd.hDevMode);
  if (!pd.hDC) {
    gputserr("PrintDlg HDC Error.");
    return ERR;
  }
  hdcViewWin=lptHDC=pd.hDC;
  hfLPT=CreateFont(   -MulDiv(fontsize, GetDeviceCaps(lptHDC, LOGPIXELSY), 72),
		       0,0,0,0, 0,0,0,0,OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,
		       DEFAULT_QUALITY,DEFAULT_PITCH,fontname);
  SelectObject(hdcViewWin,hfLPT);

  for(i=0;i<33;i++)DeleteObject(hpenViewWin[i]);
  for(i=0;i<32;i++)hpenViewWin[i]=CreatePen(PS_SOLID,MulDiv(1, GetDeviceCaps(lptHDC, LOGPIXELSY), 72),colorRef[i]);
  hpenViewWin[32]=CreatePen(PS_DOT,1,RGB(0,0,0));

  SetMapMode(lptHDC,MM_TEXT);
  x1=GetDeviceCaps(lptHDC,HORZRES);
  y1=GetDeviceCaps(lptHDC,VERTRES);
  winXCenter=x1>>1;
  winYCenter=y1>>1;
  x0=min(x1,y1);
  winXScale=winYScale=0.1*x0;
  di.cbSize=sizeof(DOCINFO);
  di.lpszDocName=(const char *)"SV Output";
  di.lpszOutput=NULL;
  LPTinit(); 
  StartDoc(lptHDC,&di);
  StartPage(lptHDC);
  return OK;
}

SV_Function void LPTclose(void)
{
  int i;
  ColorPlot=1;
  for(i=0;i<33;i++)DeleteObject(hpenViewWin[i]); /* before ginit */
  ginit();
  if (lptHDC==NULL) return;
  EndPage(lptHDC);
  EndDoc(lptHDC);
  SelectObject(hdcViewWin,hfViewWin);
  DeleteObject(hfLPT);
  DeleteDC(lptHDC);
  lptHDC=NULL;
}

SV_Function Bool EMFopen(void)
{
  int x0,x1,y1,i;

  hdcViewWin=emfHDC=CreateEnhMetaFile(NULL,filename(curwd,"sv.emf"),NULL,"SV for Windows\0sv.emf\0");
  hfEMF=CreateFont(   -MulDiv(fontsize, GetDeviceCaps(emfHDC, LOGPIXELSY), 72),
		       0,0,0,0, 0,0,0,0,OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,
		       DEFAULT_QUALITY,DEFAULT_PITCH,fontname);
  SelectObject(hdcViewWin,hfEMF);

  for(i=0;i<33;i++)DeleteObject(hpenViewWin[i]);
  for(i=0;i<32;i++)hpenViewWin[i]=CreatePen(PS_SOLID,MulDiv(1, GetDeviceCaps(emfHDC, LOGPIXELSY), 72),colorRef[i]);
  hpenViewWin[32]=CreatePen(PS_DOT,1,RGB(0,0,0));

  SetMapMode(emfHDC,MM_TEXT);
  x1=GetDeviceCaps(emfHDC,HORZRES);
  y1=GetDeviceCaps(emfHDC,VERTRES);
  winXCenter=x1>>1;
  winYCenter=y1>>1;
  x0=min(x1,y1);
  winXScale=winYScale=0.1*x0;
  return OK;
}

SV_Function void EMFclose(void)
{
  int i;
  HENHMETAFILE hMeta;
  for(i=0;i<33;i++)DeleteObject(hpenViewWin[i]); /* before ginit */
  ginit();
  if (emfHDC==NULL) return;
  SelectObject(hdcViewWin,hfViewWin);
  hMeta=CloseEnhMetaFile(emfHDC);
  DeleteEnhMetaFile(hMeta);
  DeleteObject(hfEMF);
  /*
  DeleteDC(emfHDC);
  */
  emfHDC=NULL;
}

