/* MS-Windows Initialize/Main Loop Module "winmain.c" -*- C -*- */
/*
#S#
#
# This Program is part of SV (Shelx-output Viewer)
# Copyright (C) 1991-1998 Takashi Nemoto
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details. (File "COPYING")
#
#    Note!: General Public Licence is copyrighted by the Free Software 
#           Foundation,  but this code is copyrighted by me.
#
#    Send bugs and comments to tnemoto@eels.kuicr.kyoto-u.ac.jp
#
#E#
*/

#include <windows.h>

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <mem.h>
#include "winmain.h"

#define __cmninit_c_proc__
#define __shelplot_c_proc__
#define __main_c_proc__
#define __cmd_c_proc__
#define __readf_c_proc__
#define __panic_c_proc__
#define __stack_c_proc__
#define __atom_c_proc__
#define __bond_c_proc__
#define __view_c_proc__
#define __ortep_c_proc__
#define __cavity_c_proc__
#define __message_c_proc__
#define __version_c_proc__
#define __winsub_c_proc__
#define __plotwin_c_proc__
#define __panic_c_proc__

extern HDC hdcViewWin;
extern int clsFlag;

#include "../common/def.h"

#define REGNAME "Software\\SV-Cavity\\"
#ifndef GET_X_LPARAM
#define GET_X_LPARAM(x) ((short)LOWORD(x))
#define GET_Y_LPARAM(x) ((short)HIWORD(x))
#endif 

HANDLE          MyInstance;
static char     SvwinClass[32]="SvwinClass";
static char SvwinWindowName[80]="Shelx Output Viewer for Windows";

extern void WinRefresh(void);
extern void WinPaint(void);

HWND    hwndMainWin,hwndViewWin,hwndCmdWin;
HWND    hwndLogWin,hwndListWin,hwndStackWin;
HWND    hwndListWinFrame,hwndStackWinFrame;
HFONT   hfViewWin;
extern  Float winXScale,winYScale,winXCenter,winYCenter;
Float   selectedAreaX0,selectedAreaY0,selectedAreaX1,selectedAreaY1;

char winIoBuf[256],cmdQueue[256];

TCHAR fontname[LF_FACESIZE+20];
LONG fontsize;

static BOOL FirstInstance( HANDLE );
static BOOL SetEdit2Class( HANDLE );
static BOOL AnyInstance( HANDLE, int, LPSTR );
static S_AT *getnear(Float x0,Float y0);

long FAR PASCAL WindowProc( HWND, UINT,WPARAM, LPARAM );
long FAR PASCAL MainWindowProc( HWND, UINT , WPARAM, LPARAM );
long FAR PASCAL ViewWindowProc( HWND, UINT, WPARAM, LPARAM );
long FAR PASCAL Edit2WindowProc( HWND, UINT, WPARAM, LPARAM);
long FAR PASCAL CmdWindowProc( HWND, UINT , WPARAM, LPARAM );
BOOL FAR PASCAL  ListWinFrameProc(HWND,UINT,WPARAM,LPARAM);
BOOL FAR PASCAL  StackWinFrameProc(HWND,UINT,WPARAM,LPARAM);
WNDPROC OrgEditWindowProc;
void ViewWinFontSelect(void);

#include <direct.h>

void CommandExecute(char *cmd) {
  sprintf(cmdQueue,"%s",cmd);
  PostMessage(hwndMainWin,WM_COMMAND,COMMAND_EXEC,0);
}

/*
 * WinMain - initialization, message loop
 */
int PASCAL WinMain( HINSTANCE this_inst, HINSTANCE prev_inst, LPSTR cmdline,
		   int cmdshow )
{
  MSG         msg;
  HACCEL haccel;
  char DummyCommandLine[cBufSize];
  char **argv;

  MAX_LINE=LIST_WINDOW_LINES;
  MyInstance = this_inst;
  if( !prev_inst ) {
    if( !FirstInstance( this_inst ) ) return( FALSE );
  }
  if ( ! SetEdit2Class( this_inst )) return (FALSE);

  if( !AnyInstance( this_inst, cmdshow, cmdline ) ) return( FALSE );
  argv=(char **)malloc(sizeof(char *)*2);
  GetWindowsDirectory(DummyCommandLine,cBufSize);
  strcat(DummyCommandLine,"\\dummy.exe");
  argv[0]=(char *)malloc(strlen(DummyCommandLine)+1);
  argv[1]=cmdline;
  CommonInitialize();
  /*
  PrintVer();
  PrintMess();
  */
  /*   InitMain(2,argv); */
  WinRefresh();
  SetFocus(hwndCmdWin);
  haccel=LoadAccelerators(MyInstance,"SvwinAccel");
  ginit();

  if (argv[1]!=NULL && argv[1][0]!=0) {
    sprintf(fbuf,"openfile %s",argv[1]);
    AppendCommandQueue(fbuf);
  }


/*  InitCommonControls(); */
  while( GetMessage( &msg, NULL, 0, 0 ) ) {
    if (!TranslateAccelerator(
			      hwndMainWin,haccel,&msg)) 
      {
	TranslateMessage( &msg );
	DispatchMessage( &msg );
      }
    if (ExecCommand()<0) break;
  }
  return( msg.wParam );
} /* WinMain */

/*
 * FirstInstance - register window class for the application,
 *                 and do any other application initialization
 */
static BOOL FirstInstance( HANDLE this_inst )
{
  WNDCLASS    wc;
  BOOL        rc;

  /*
   * set up and register window class
   */
  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc = (LPVOID) WindowProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = sizeof( DWORD );
  wc.hInstance = this_inst;
  wc.hIcon = LoadIcon( this_inst, "SvwinIcon" );
  if (wc.hIcon==NULL) return 0;
  wc.hCursor = LoadCursor( NULL, IDC_ARROW );
  wc.hbrBackground = GetStockObject( WHITE_BRUSH );
  wc.lpszMenuName = "SvwinMenu";
  wc.lpszClassName = SvwinClass;
  rc = RegisterClass( &wc );
  return( rc );

} /* FirstInstance */

static BOOL SetEdit2Class( HANDLE this_inst )
{
  WNDCLASS    wc;
  BOOL        rc;
  rc=GetClassInfo(NULL,"EDIT",&wc);
  OrgEditWindowProc=wc.lpfnWndProc; 
  wc.lpfnWndProc = (LPVOID) Edit2WindowProc;
  wc.lpszMenuName = NULL;
  wc.hInstance = this_inst;
  wc.lpszClassName = "EDIT2";
  rc = RegisterClass( &wc );
  return rc;
}

void SaveFontSetting() {
    HKEY hk;
    DWORD tp;
    if (RegOpenKeyEx(HKEY_CURRENT_USER, REGNAME, 
		     0, KEY_ALL_ACCESS, &hk)==ERROR_SUCCESS ||
	RegCreateKeyEx(HKEY_CURRENT_USER, REGNAME,
		       0, NULL, 0, KEY_ALL_ACCESS, 
		       NULL, &hk, &tp)==ERROR_SUCCESS) {
      tp=fontsize;
      RegSetValueEx(hk, "FontSize",
		    0, REG_DWORD, &tp, 4);
      RegSetValueEx(hk, "FontName",
		    0, REG_SZ, fontname, strlen(fontname)+1);
      RegCloseKey(hk);
    }
}
void LoadFontSetting() {
  DWORD len,tp,fs;
  HKEY hk;
  char buf[LF_FACESIZE+10];

  sprintf(fontname,"Courier");
  fontsize=10;

  if (RegOpenKeyEx(HKEY_CURRENT_USER, REGNAME, 
		   0, KEY_ALL_ACCESS, &hk)==ERROR_SUCCESS) {
    
    len=LF_FACESIZE;
    if (RegQueryValueEx(hk,"FontName",
			NULL,NULL,buf,&len)==ERROR_SUCCESS) {
      strcpy(fontname,buf);
    } 
    len=4;
    if (RegQueryValueEx(hk,"FontSize",
			NULL,NULL,&fs,&len)==ERROR_SUCCESS) {
      fontsize=fs;
    }
    RegCloseKey(hk);
  }
}    

static BOOL AnyInstance( HANDLE this_inst, int cmdshow, LPSTR cmdline )
{
  FARPROC proc;
  HMENU listWinMenu;
  hwndMainWin = CreateWindowEx(WS_EX_ACCEPTFILES, 
			       SvwinClass,SvwinWindowName,WS_OVERLAPPEDWINDOW,
			       /*       50,50,350,380,*/
			       CW_USEDEFAULT,CW_USEDEFAULT,380,380,
			       NULL,NULL,this_inst,NULL);    
  if( !hwndMainWin ) return(0);

  proc=MakeProcInstance(ListWinFrameProc,this_inst);
  hwndListWinFrame=CreateDialog(this_inst,"ListWinFrame",hwndMainWin,proc);
  hwndListWin=CreateWindow("EDIT","ListWin",
			   WS_CHILD|WS_BORDER|ES_MULTILINE|ES_AUTOVSCROLL|
			   WS_VSCROLL,
			   0,0,350,350,hwndListWinFrame,NULL,this_inst,NULL);
  listWinMenu=LoadMenu(MyInstance,"ListWinMenu");
  SetMenu(hwndListWin,listWinMenu);
  proc=MakeProcInstance(StackWinFrameProc,this_inst);
  hwndStackWinFrame=CreateDialog(this_inst,"StackWinFrame",hwndMainWin,proc);
  hwndStackWin=CreateWindow("EDIT","StackWin",
			    WS_CHILD|WS_BORDER|ES_MULTILINE|ES_AUTOVSCROLL|
			    WS_VSCROLL,
			    0,0,350,350,hwndStackWinFrame,NULL,this_inst,NULL);
  /*   FreeProcInstance(proc);*/
  if( !hwndListWinFrame ) return(0);

  MoveWindow(hwndMainWin,20,20,400,450,TRUE);
  MoveWindow(hwndListWinFrame,420,20,300,300,TRUE);
  MoveWindow(hwndStackWinFrame,420,320,300,150,TRUE);

  hwndCmdWin=CreateWindow("EDIT2","CmdWin",WS_CHILD |WS_BORDER,
			  0,0,350,30,hwndMainWin,NULL,this_inst,NULL);
  hwndViewWin=CreateWindow(SvwinClass,"ViewWin",
			   WS_CHILD | WS_BORDER,0,30,350,350,
			   hwndMainWin,NULL,this_inst,NULL);
  ShowWindow( hwndMainWin, cmdshow );
  UpdateWindow( hwndMainWin ); 
  ShowWindow(hwndViewWin,cmdshow);
  UpdateWindow(hwndViewWin);                       
  ShowWindow(hwndCmdWin,cmdshow);
  UpdateWindow(hwndCmdWin);
  ShowWindow(hwndListWin,cmdshow);
  UpdateWindow(hwndListWin);
  ShowWindow(hwndStackWin,cmdshow);
  UpdateWindow(hwndStackWin);

  gopen();

  LoadFontSetting();
  SaveFontSetting();

  hfViewWin=CreateFont(fontsize,
		       0,0,0,0, 0,0,0,0,OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,
		       DEFAULT_QUALITY,DEFAULT_PITCH,fontname);

  SendMessage(hwndListWin,WM_SETFONT,(WPARAM)hfViewWin,1);
  SendMessage(hwndStackWin,WM_SETFONT,(WPARAM)hfViewWin,1);
  SendMessage(hwndCmdWin,WM_SETFONT,(WPARAM)hfViewWin,1);
  SelectObject(hdcViewWin,hfViewWin);
  SendMessage(hwndCmdWin, EM_SETSEL, 0, -1);
  SendMessage(hwndCmdWin,WM_CLEAR, 0, 0L);

  SendMessage(hwndListWin,EM_SETLIMITTEXT,(WPARAM)10000,0);
  SendMessage(hwndStackWin,EM_SETLIMITTEXT,(WPARAM)10000,0);
    
  return( TRUE );
                        
} /* AnyInstance */


/*
 * AboutDlgProc - processes messages for the about dialog.
 */
BOOL PASCAL AboutDlgProc( HWND hwnd, unsigned msg,
			 UINT wparam, LONG lparam )
{
  lparam = lparam;		/* turn off warning */

  switch( msg ) {
  case WM_INITDIALOG:
    return( TRUE );
  case WM_COMMAND:
    if( LOWORD( wparam ) == IDOK ) {
      EndDialog( hwnd, TRUE );
      return( TRUE );
    }
    break;
  }
  return( FALSE );

} /* AboutDlgProc */

/*
 * WindowProc - handle messages for the main application window
 */
LONG PASCAL WindowProc( HWND hwnd, unsigned msg,
		       UINT wparam, LONG lparam )
{
  /* Common CallBack Procedure */
  switch(msg){
  case WM_CREATE:
    DragAcceptFiles(hwnd,1);
    break;
  case WM_DROPFILES:
    {
      char buf[cBufSize];
      SetForegroundWindow(hwndMainWin);
      SetFocus(hwndCmdWin);
      DragQueryFile((HDROP)wparam,0,buf,cBufSize-1);
      svx(buf);
      refresh();
      DragFinish((HDROP)wparam);
    }    
    return 0;
  default:
    break;
  }
  /* */
  if (hwnd==hwndViewWin) {
    return ViewWindowProc(hwnd,msg,wparam,lparam);
  } else if (hwnd==hwndMainWin) {
    return MainWindowProc(hwnd,msg,wparam,lparam);
  } else if (hwnd==hwndCmdWin){
    return CmdWindowProc(hwnd,msg,wparam,lparam);
  }
  return( DefWindowProc( hwnd, msg, wparam, lparam ) );
}

LONG PASCAL ListWindowProc( HWND hwnd, unsigned msg,
			   UINT wparam, LONG lparam )
{
  return( DefWindowProc( hwnd, msg, wparam, lparam ) );
}

void WinAtExitProc(void) {
  PostQuitMessage(0);
}

LONG PASCAL MainWindowProc( HWND hwnd, unsigned msg,
			   UINT wparam, LONG lparam ) {
  FARPROC     proc;
  char *fname;
  static int lastWindowPosX=-1000,lastWindowPosY=-1000;
  switch( msg ) {
  case WM_COMMAND:
    switch( LOWORD( wparam ) ) {
    case MENU_ABOUT:
      /*
      proc = MakeProcInstance( AboutDlgProc, MyInstance );
      DialogBox( MyInstance,"AboutBox", hwnd, proc );
      FreeProcInstance( proc );
      */
      PrintMess();
      break;
    case MENU_OPEN:
      fname=InputFileName("Open New File","shelxl.res",1);
      if (fname!=NULL) {
	sprintf(fbuf,"openfile %s",fname);
	AppendCommandQueue(fbuf);
      }
      break;
    case MENU_PLOT_MONO:
      AppendCommandQueue("plot");
      break;
    case MENU_PLOT_COLOR:
      AppendCommandQueue("plot color");
      break;

    case MENU_EXPORT_CAVITY:
      AppendCommandQueue("export cavity");
      break;
    case MENU_EXPORT_ORTEP:
      AppendCommandQueue("export ortep");
      break;
    case MENU_EXPORT_CSD:
      AppendCommandQueue("export csd");
      break;
    case MENU_EXPORT_CHARON:
      AppendCommandQueue("export charon");
      break;
    case MENU_EXPORT_SHELXL:
      AppendCommandQueue("export shelxl");
      break;
    case MENU_EXPORT_CHEMDRAW:
      AppendCommandQueue("export chemdraw");
      break;
    case MENU_EXPORT_RASMOL:
      AppendCommandQueue("export rasmol");
      break;
    case MENU_EXPORT_MACTEMPAS:
      AppendCommandQueue("export mactempas");
      break;
    case MENU_EXPORT_POVRAY:
      AppendCommandQueue("export povray");
      break;

    case MENU_EXIT:
      WinAtExitProc();
      break;
    case MENU_FONTSEL:
      ViewWinFontSelect();
      break;
    case COMMAND_EXEC:
      AppendCommandQueue(cmdQueue);
      break;
    case MENU_UNITCELL_ON:
      CommandExecute("in cell");
      break;
    case MENU_UNITCELL_OFF:
      CommandExecute("om cell");
      break;
    case MENU_BONDLENGTH_ON:
      CommandExecute("in bl");
      break;
    case MENU_BONDLENGTH_OFF:
      CommandExecute("om bl");
      break;
    case MENU_PROJECTION_A:
      CommandExecute("vw a");
      break;
    case MENU_PROJECTION_B:
      CommandExecute("vw b");
      break;
    case MENU_PROJECTION_C:
      CommandExecute("vw c");
      break;
    case MENU_LABEL_LONG:
      CommandExecute("in ll");
      break;
    case MENU_LABEL_SHORT:
      global_flag &= ~GLOBALFLAG_LONGLABEL;
      CommandExecute("in lb");
      break;
    case MENU_LABEL_OFF:
      CommandExecute("om lb");
      break;
    case MENU_STEREO_ON:
      CommandExecute("st");
      break;
    case MENU_STEREO_OFF:
      CommandExecute("mo");
      break;
    case MENU_AREA_INSIDE:
      if (clsFlag==0){
	char buf[1024];
	sprintf(buf,"selectrectin %f %f %f %f",selectedAreaX0,selectedAreaY0,
		selectedAreaX1,selectedAreaY1);
	CommandExecute(buf);
      };
      break;
    case MENU_AREA_OUTSIDE:
      if (clsFlag==0){
	char buf[1024];
	sprintf(buf,"selectrectout %f %f %f %f",selectedAreaX0,selectedAreaY0,
		selectedAreaX1,selectedAreaY1);
	CommandExecute(buf);
      };
      break;
    case MENU_MAKEBOND:
      if (clsFlag==0){
	char buf[1024];
	S_AT *a,*b;
	a=getnear(selectedAreaX0,selectedAreaY0);
	b=getnear(selectedAreaX1,selectedAreaY1);
	if (a!=NULL && b!=NULL && a!=b){
	  sprintf(buf,"mb #%d #%d",a->num,b->num);
	  CommandExecute(buf);
	}
      };
      break;
    case MENU_DELBOND:
      if (clsFlag==0){
	char buf[1024];
	S_AT *a,*b;
	a=getnear(selectedAreaX0,selectedAreaY0);
	b=getnear(selectedAreaX1,selectedAreaY1);
	if (a!=NULL && b!=NULL && a!=b){
	  sprintf(buf,"db #%d #%d",a->num,b->num);
	  CommandExecute(buf);
	}
      };
      break;
    case MENU_MOVEATOM:
      if (clsFlag==0){
	char buf[1024];
	S_AT *a,*b;
	b=getnear(selectedAreaX0,selectedAreaY0);
	a=getnear(selectedAreaX1,selectedAreaY1);
	if (a!=NULL && b!=NULL && a!=b){
	  sprintf(buf,"rb #%d #%d",a->num,b->num);
	  CommandExecute(buf);
	}
      };
      break;
    case MENU_DUPATOM:
      if (clsFlag==0){
	char buf[1024];
	S_AT *a,*b;
	b=getnear(selectedAreaX0,selectedAreaY0);
	a=getnear(selectedAreaX1,selectedAreaY1);
	if (a!=NULL && b!=NULL && a!=b){
	  sprintf(buf,"du #%d #%d",a->num,b->num);
	  CommandExecute(buf);
	}
      };
      break;
#ifdef PANIC
    case MENU_PANIC:
      CommandExecute("panic");
      break;
#endif
    default:
      return( DefWindowProc( hwnd, msg, wparam, lparam ) );
    }
    break;
  case WM_DESTROY:
    WinAtExitProc();
    break;
  case WM_SIZE:
    {
      RECT rect;
      GetClientRect(hwndMainWin,&rect);
      MoveWindow(hwndCmdWin,0,0,rect.right,29,TRUE);
      MoveWindow(hwndViewWin,0,30,rect.right,rect.bottom-30,TRUE);
    }
    break;
   
  case WM_WINDOWPOSCHANGING:
    {
      RECT rect;
      GetWindowRect(hwndMainWin,&rect);
      lastWindowPosX=rect.right;
      lastWindowPosY=rect.top;
    }
  case WM_WINDOWPOSCHANGED:
    {
      RECT rect,rect2;
      int dx,dy,xPos,yPos;
      GetWindowRect(hwndMainWin,&rect);
      dx=rect.right-lastWindowPosX;
      dy=rect.top-lastWindowPosY;
      
      GetWindowRect(hwndStackWinFrame,&rect);
      rect.left+=dx; rect.right+=dx;
      rect.bottom+=dy; rect.top+=dy;
      MoveWindow(hwndStackWinFrame,rect.left,rect.top,rect.right-rect.left,rect.bottom-rect.top,TRUE);
      GetWindowRect(hwndListWinFrame,&rect);
      rect.left+=dx; rect.right+=dx;
      rect.bottom+=dy; rect.top+=dy;
      MoveWindow(hwndListWinFrame,rect.left,rect.top,rect.right-rect.left,rect.bottom-rect.top,TRUE);
    }
  default:
    return( DefWindowProc( hwnd, msg, wparam, lparam ) );
  }
  return 0;
}


void ViewWinFontSelect(void) {
  LOGFONT lf;
  CHOOSEFONT cf;
  int xx;
  DeleteObject(hfViewWin);
  memset(&cf, 0, sizeof(CHOOSEFONT));
  cf.lStructSize = sizeof(CHOOSEFONT);
  cf.hwndOwner = hwndMainWin;
  cf.lpLogFont = &lf;
  cf.Flags = CF_SCREENFONTS | CF_EFFECTS;
  cf.rgbColors = RGB(0, 0, 0); 
  cf.nFontType = SCREEN_FONTTYPE;
  memcpy(lf.lfFaceName,fontname,sizeof(fontname));
  xx=ChooseFont(&cf);
  /*
  hfViewWin = CreateFontIndirect(cf.lpLogFont);
  */

  memcpy(fontname,lf.lfFaceName,sizeof(fontname));
  fontsize=lf.lfHeight;

  hfViewWin=CreateFont(fontsize,
		       0,0,0,0, 0,0,0,0,OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,
		       DEFAULT_QUALITY,DEFAULT_PITCH,fontname);
  SaveFontSetting();


  SendMessage(hwndListWin,WM_SETFONT,(WPARAM)hfViewWin,1);
  SendMessage(hwndStackWin,WM_SETFONT,(WPARAM)hfViewWin,1);
  SendMessage(hwndCmdWin,WM_SETFONT,(WPARAM)hfViewWin,1);
  SelectObject(hdcViewWin,hfViewWin);
  WinRefresh();
}

LONG PASCAL CmdWindowProc( HWND hwnd, unsigned msg,
			  UINT wparam, LONG lparam )
{
  return( DefWindowProc( hwnd, msg, wparam, lparam ) );
}

LONG PASCAL Edit2WindowProc( HWND hwnd, UINT msg,
			    WPARAM wparam, LPARAM lparam )
{
  LONG rc;
  switch( msg ) {
  case  WM_CHAR:
    if (wparam==13) {
      WORD cbText;

      *(WORD *) cmdQueue = sizeof(cmdQueue) - 1; 
      cbText = SendMessage(hwnd,EM_GETLINE,
			   0,(LPARAM) cmdQueue);     
      cmdQueue[cbText] = '\0';
      SendMessage(hwndCmdWin, EM_SETSEL, 0, -1);
      SendMessage(hwndCmdWin,WM_CLEAR, 0, 0L);
      PostMessage(hwndMainWin, WM_COMMAND, COMMAND_EXEC,0); 
      return 0;
    }
  }
  rc=CallWindowProc(OrgEditWindowProc, hwnd, msg, wparam, lparam ) ;
  return rc;
}

static void DrawRect(int xa,int ya,int xb,int yb){
  int tmp;
  RECT rect;
  S_AT *a,*b;
  rect.left=min(xa,xb);
  rect.top=min(ya,yb);
  rect.right=max(xa,xb);
  rect.bottom=max(ya,yb);
  gopen();
  DrawFocusRect(hdcViewWin,&rect);
  a=getnear(((Float)xa-winXCenter)/winXScale,
	    (winYCenter-(Float)ya)/winYScale);
  b=getnear(((Float)xb-winXCenter)/winXScale,
	    (winYCenter-(Float)yb)/winYScale);
  if (a!=NULL && b!=NULL){
    tmp=SetROP2(hdcViewWin,R2_XORPEN);
    gcolor(32);
    gline(a->axy[0],a->axy[1],b->axy[0],b->axy[1]);
    SetROP2(hdcViewWin,tmp);
  }
  gclose();
};

LONG PASCAL ViewWindowProc( HWND hwnd, unsigned msg,
			   UINT wparam, LONG lparam ) {
  static int xa,ya,xb,yb,x0,y0,gf,button=FALSE;
  static S_AT *lastAtom;

  switch( msg ) {
  case  WM_PAINT:
    WinPaint();
    break;
  case WM_CREATE:
    WinRefresh();
    break;

  case WM_RBUTTONDOWN:
  case WM_MBUTTONDOWN:
    xa = GET_X_LPARAM(lparam); 
    ya = GET_Y_LPARAM(lparam);
    button = TRUE; 
    break;
  case WM_RBUTTONUP:
  case WM_MBUTTONUP:
    sprintf(fbuf,"setvmt %f %f %f %f %f %f %f %f %f %f %f\n",
	    VMT[0][0],VMT[0][1],VMT[0][2],
	    VMT[1][0],VMT[1][1],VMT[1][2],
	    VMT[2][0],VMT[2][1],VMT[2][2],
	    shift[0],shift[1]
	    );
    AppendCommandQueue(fbuf);
    button = FALSE;
    SetFocus(hwndCmdWin);
    refresh();
    break;
  case WM_LBUTTONDOWN:
    if (clsFlag==0){
      DrawRect(xa,ya,x0,y0);
    }
    x0 = xa =xb= GET_X_LPARAM(lparam); 
    y0 = ya =yb= GET_Y_LPARAM(lparam);
    DrawRect(xa,ya,xb,yb);
    clsFlag=0;
    button = TRUE; 
    break;
  case WM_LBUTTONUP:
    if ((xa-xb)*(xa-xb)+(ya-yb)*(ya-yb)<=9.1) {
      S_AT *a;
      a=getnear(((Float)xb-winXCenter)/winXScale,
		(winYCenter-(Float)yb)/winYScale);
      if (a!=NULL) {
	if (GetKeyState(VK_CONTROL)&0x8000){
	  sprintf(fbuf,"select fragment #%d",a->num);
	} else {
	  sprintf(fbuf,"select #%d",a->num);
	}
	AppendCommandQueue(fbuf);
	lastAtom=a;
      } else {
	lastAtom=NULL;
      }
    } else {
      selectedAreaX0=((Float)xa-winXCenter)/winXScale;
      selectedAreaY0=(winYCenter-(Float)ya)/winYScale;
      selectedAreaX1=((Float)xb-winXCenter)/winXScale;
      selectedAreaY1=(winYCenter-(Float)yb)/winYScale;
    }
    SetFocus(hwndCmdWin);
    break;
  case WM_MOUSEMOVE:
    if (button == FALSE) break;
    xb = GET_X_LPARAM(lparam);
    yb = GET_Y_LPARAM(lparam);
    if ((wparam & MK_MBUTTON) ||
	((wparam & MK_RBUTTON) &&
	 (wparam & (MK_LBUTTON|MK_SHIFT)))){
      gf=global_flag;
      global_flag&=~GLOBALFLAG_TMP_LABEL;
      shift[0]-=((Float)xb-xa)/winXScale/vscale;
      shift[1]+=((Float)yb-ya)/winYScale/vscale;
      refresh();
      global_flag=gf;
      xa=xb;
      ya=yb;
                            
    } else if (wparam & MK_RBUTTON) {
      int gf;
      Double dx,dy,px,py,r,r2,theta,theta2;
      gf=global_flag;
      global_flag&=~GLOBALFLAG_TMP_LABEL;
      dx=xb-winXCenter;
      dy=yb-winYCenter;
      px=xa-winXCenter;
      py=ya-winYCenter;
      xa=xb;
      ya=yb;
      r2=sqrt(dx*dx+dy*dy);
      if (r2>winXScale*1.5)  {
	r=sqrt(px*px+py*py);
	theta=atan2(py,px);
	theta2=atan2(dy,dx);
	px=dx*(r-r2)/r2;
	py=dy*(r-r2)/r2;
	rotx(PI*py*0.1/winXScale);
	roty(PI*px*0.1/winYScale);
	rotz(theta-theta2);
      } else {
	rotx(PI*(py-dy)*0.1/winXScale);
	roty(PI*(px-dx)*0.1/winYScale);
      }
      refresh();
      global_flag=gf;
    } else if (wparam & MK_LBUTTON) {
      DrawRect(xa,ya,x0,y0);
      x0=xb;
      y0=yb;
      DrawRect(xa,ya,x0,y0);
    }
    break;
  default:
    return( DefWindowProc( hwnd, msg, wparam, lparam ) );
  }
  return( 0L );
};

S_AT *getnear(Float x0,Float y0)
{
  Double xx,yy,len,minlen;
  int i,num=0;
  minlen=10000;
  for (i=0;i<n_satom;i++)
    {
      xx=SAT[i].axy[0]-x0;
      yy=SAT[i].axy[1]-y0;
      len=xx*xx+yy*yy;
      if (len<minlen && (SAT[i].view_flag & FLAG_OMIT)==0) 
        {
          minlen=len;
          num=i;
        }
    }
  /*  if (minlen>10) return NULL;*/
  return SAT+num;
}



BOOL PASCAL  ListWinFrameProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
  switch(msg) {
  case WM_SIZE:
    {
      RECT rect;
      GetClientRect(hwnd,&rect);
      MoveWindow(hwndListWin,0,0,rect.right,rect.bottom,TRUE);
      return 1;
    }
  }
  return 0;
}

BOOL PASCAL  StackWinFrameProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
  switch(msg) {
  case WM_COMMAND:
    switch(LOWORD(wparam)) {
    case STACKWIN_MENU_COPY:
      PostMessage(hwndStackWin,WM_COPY,0,0);
      return 1;
    case STACKWIN_MENU_OMIT:
      CommandExecute("om sl");
      return 1;
    case STACKWIN_MENU_INCLUDE:
      CommandExecute("in sl");
      return 1;
    case STACKWIN_MENU_CLEAR:
      CommandExecute("cs");
      return 1;
    case STACKWIN_MENU_POP:
      popstack();
      refresh();
      return 1;
    case STACKWIN_MENU_RESTORE:
      CommandExecute("rs");
      return 1;
    case STACKWIN_MENU_MAKEBOND:
      CommandExecute("mb sl");
      return 1;
    case STACKWIN_MENU_MBPAIR:
      CommandExecute("mb pair");
      return 1;
    case STACKWIN_MENU_REMBOND:
      CommandExecute("db sl");
      return 1;
    case STACKWIN_MENU_RBPAIR:
      CommandExecute("db pair");
      return 1;
    }
    break;    
  case WM_SIZE:
    {
      RECT rect;
      GetClientRect(hwnd,&rect);
      MoveWindow(hwndStackWin,0,0,rect.right,rect.bottom,TRUE);
      return 1;
    }
  }
  return 0;
}
