/* Subroutines for MS-Win Module "winsub.c" -*- C -*- */

/*
#S#
#
# This Program is part of SV (Shelx-output Viewer)
# Copyright (C) 1991-1998 Takashi Nemoto
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details. (File "COPYING")
#
#    Note!: General Public Licence is copyrighted by the Free Software 
#           Foundation,  but this code is copyrighted by me.
#
#    Send bugs and comments to tnemoto@eels.kuicr.kyoto-u.ac.jp
#
#E#
*/

#include <windows.h>

#define __panic_c__
#define __winsub_c__
#define __winmain_c_proc__

#include "../common/def.h"

SV_Function void *Malloc(int size)
{
  void *X;
  X=GlobalAlloc(GMEM_FIXED,size);
  if (X==0) {
        FatalError("Out of Memory!");
  }
  return X;
}

SV_Function void *Calloc(int nMember,int size)
{
  void *X;
  X=GlobalAlloc(GMEM_ZEROINIT|GMEM_FIXED,nMember*size);
  if (X==0) {
        FatalError("Out of Memory!");
  }
  return X;
}

SV_Function void Free(void *pt)
{
  GlobalFree(pt);
}

SV_Function void *Realloc(void *pt,int *nMemb,int deltaMemb,int s)
{
  HGLOBAL pt1;
  int size,newSize;
  size=s*(*nMemb);
  (*nMemb)+=deltaMemb*s;
  newSize=(*nMemb)*s;
  pt1=GlobalReAlloc(pt,newSize,0); 
  if (pt1==NULL) {
    pt1=GlobalAlloc(GMEM_FIXED,newSize);
    if (pt1==NULL) {
      FatalError("Out of Memory");
    }
#ifdef DEBUG
    {
      char buff[256];
      sprintf(buff,"Size %d\n",size);
      PutMessage1(buff);
      sprintf(buff,"NewSize %d\n",newSize);
      PutMessage1(buff);
    }
#endif
    if (size>newSize) size=newSize;
    if (size>0) {
      memcpy(pt1,pt,size);
    }
  }
  return pt1;
}


SV_Function int *SelectItem(char *prompt,int n_item,char **ItemList)
{
        return 0;
}

SV_Function char *InputFileName(char *prompt,char *filename,int isInput)
{
    static char FileName[cBufSize];
    int flag;
    OPENFILENAME ofn;
    char FilterString[]="All\0*.*\0"
        "ChemOffice\0*.ct;*.con\0"
        "Bitmap\0*.bmp;*.gif;*.mag;*.pic\0"
        "CSD(FDAT)\0*.dat\0\0\0";
    strcpy(FileName,filename);
    ofn.lStructSize=sizeof(OPENFILENAME);
    ofn.hwndOwner=NULL;
    ofn.hInstance=NULL;
    ofn.lpstrFilter=FilterString;
    ofn.lpstrCustomFilter=NULL;
    ofn.nMaxCustFilter=0;
    ofn.nFilterIndex=0;
    ofn.lpstrFile=FileName;
    ofn.nMaxFile=cBufSize;
    ofn.lpstrFileTitle=NULL;
    ofn.nMaxFileTitle=0;
    ofn.lpstrInitialDir=curwd;
    ofn.lpstrTitle=prompt;
    ofn.Flags=OFN_HIDEREADONLY | OFN_NOCHANGEDIR;
    ofn.nFileOffset=0;
    ofn.nFileExtension=0;
    ofn.lpstrDefExt=NULL;
/*  ofn.lCustData=;*/
/*  ofn.lpfnHook;*/
/*  ofn.lpTemplateName*/    
    if (isInput) {
        flag=GetOpenFileName(&ofn);
        if (flag==FALSE) return NULL;   
        return FileName;
    } else {
        ofn.Flags|=OFN_CREATEPROMPT|OFN_OVERWRITEPROMPT;
        flag=GetSaveFileName(&ofn);
        if (flag==FALSE) return NULL;    
        return FileName;
    }
}

SV_GlobalFunction void PutMessage(char *st)
{
    #define MessageBufferLength 2000
    static char MessageBuf[MessageBufferLength];
    static int len=0;
    static int lines=0;
    if (strcmp("",st)==0){
        MessageBuf[0]=0;
        len=0;
        lines=0;
    } else if (strcmp(".",st)==0) {
        gputsinfo(MessageBuf);
        PutMessage("");
    } else {
        if (strlen(st)+len < MessageBufferLength-2) {
           sprintf(MessageBuf+len,"%s\n",st);
           lines++;
           len=strlen(MessageBuf);
        }
    }
}

SV_GlobalFunction void PutMessage1(char *X)
{
    PutMessage("");
    PutMessage(X);
    PutMessage(".");
}

SV_GlobalFunction void Exit(int dummy)
{
    WinAtExitProc();
}

SV_GlobalFunction void PutWarning1(char *X)
{
    PutMessage("");
    PutMessage(X);
    PutMessage(".");
}

SV_GlobalFunction void FatalError(char *X)
{
    PutMessage("");
    PutMessage(X);
    PutMessage(".");
    Exit(1);
}

SV_GlobalFunction void ClearMessage(void)
{
}

SV_Function SetWaitCursor(void)
{
}

SV_Function UnSetWaitCursor(void)
{
}

