/* Unix Depend Subroutines "unixsub.c" -*- C -*- */
/*
#S#
#
# This Program is part of SV (Shelx-output Viewer)
# Copyright (C) 1991-1998 Takashi Nemoto
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details. (File "COPYING")
#
#    Note!: General Public Licence is copyrighted by the Free Software 
#           Foundation,  but this code is copyrighted by me.
#
#    Send bugs and comments to tnemoto@eels.kuicr.kyoto-u.ac.jp
#
#E#
*/

#define __unixsub_c__
#define __xtcalls_c_proc__
#include "def.h"

#ifdef DEBUG
void *PMallocList[1024];
int MallocSizeList[1024];
int MallocCount=0;

void MallocList(void* X,int size){
  fprintf(stderr,"Malloc: %lx (%d)\n",X,size);
  PMallocList[MallocCount]=X;
  MallocSizeList[MallocCount++]=size;
};

void FreeList(void* ptr){
  int i,j;
  if (ptr==NULL) return;
  for(i=0;i<MallocCount;i++){
    if (ptr==PMallocList[i]){
      fprintf(stderr,"Free: %lx (%d)\n",ptr,MallocSizeList[i]);
      MallocCount--;
      for(j=i;j<MallocCount;j++){
	PMallocList[j]=PMallocList[j+1];
	MallocSizeList[j]=MallocSizeList[j+1];
      }
      return;
    }
  }
  fprintf(stderr,"Illegal Free!\n");
}
#else
#define MallocList(a,b)
#define FreeList(a)
#endif


SV_GlobalFunction void *Malloc(int size)
{
  char *X;
  X=malloc(size);
  if (X==NULL) {
    FatalError("Out of Memory!!");
  }
  MallocList(X,size);
  return X;
}

SV_GlobalFunction void *Calloc(int nmemb,int size)
{
  char *X;
  X=calloc(nmemb,size);
  if (X==NULL) {
    FatalError("Out of Memory!!");
  }
  /******************************************/
  MallocList(X,size);

  return X;
}

SV_GlobalFunction void Free(void *X)
{
  if (X==NULL) return;
  FreeList(X);
  free(X);
}

SV_GlobalFunction void *Realloc(void *org,int *oldSize,int deltaMemb,int size)
{
  char *X;
  int oSize,newSize;
  oSize=size*(*oldSize);
  newSize=oSize+deltaMemb*size;
  (*oldSize)+=deltaMemb;
  /******************************************/
  FreeList(org);

  X=realloc(org,newSize);

  /******************************************/
  MallocList(X,newSize);

  if (X==NULL) {
    X=malloc(newSize);
    /******************************************/
    MallocList(X,newSize);

    if (X==NULL) {
      FatalError("Out of Memory!!");
    }
    size=oSize;
    if (size>newSize) size=newSize;
    if (org!=NULL){
      memcpy(X,org,size);
    }
    /******************************************/
    FreeList(org);
    free(org);
  }
/*fprintf(stderr,"Realloc Out\n");*/
  return X;
}

SV_GlobalFunction void FatalError(char *X)
{
  fprintf(stderr,"Fatal Error\n");
  fprintf(stderr,"%s\n",X);
  Exit(1);
}

SV_GlobalFunction void PutMessage(char *X)
{
    if (strcmp(X,"")!=0 && strcmp(X,".")!=0) fprintf(stderr,"%s\n",X);
}

SV_GlobalFunction void PutMessage1(char *X)
{
    PutMessage("");
    PutMessage(X);
    PutMessage(".");
}

SV_GlobalFunction void PutWarning1(char *X)
{
    PutMessage("");
    PutMessage(X);
    PutMessage(".");
}

SV_GlobalFunction void Exit(int i)
{
    exit(i);
}


SV_Function char *InputFileName(char *prompt,char *fileName,int IsRead)
{
  static char fname[cBufSize];
  char prmpt[cBufSize];
  strcpy(fname,fileName);
  sprintf(prmpt,"%s - Input File Name  :",prompt);
  if (X11MODE) {
    if (KeyInput(prmpt,fname,fname,cBufSize)<=0) {
      return NULL;
    }
  } else {
    fprintf(stderr,"Input File Name: ");
    fgets(fname,cBufSize,stdin);
  }
  return fname;
}

SV_Function int SelectItem(char *Prompt,int n_data,char **datalist)
{
  return 0;
}

SV_Function void ChangeWallPaper(char *type,char *FileName)
{
  char *fbuffer;
  if (X11MODE) {
    sprintf(fbuffer,"xv -root -quit %s",FileName);
    system(fbuffer);
  }
}

SV_Function int prompt_input(char *mess,char *dflt,char *destin,int len)
{
  if (X11MODE) {
    return KeyInput(mess,dflt,destin,len);
  }
  fprintf(stderr,"%s\n",mess);
  fgets(destin,len,stdin);
  return 1;
}

SV_Function int ClearMessage(void)
{
  if (X11MODE) ErrorPopdown();
  return OK;
}

SV_Function int SetWaitCursor(void)
{
  XSetWaitCursor();
  return OK;
}

SV_Function int UnSetWaitCursor(void)
{
  XUnsetWaitCursor();
  return OK;
}
